#!/usr/bin/env python
"""
.. module:: pymeteo.skewt
   :platform: Unix, Windows
   :synopsis: Skew-T/Log-P plotting

.. moduleauthor:: Casey Webster <casey.webster@gmail.com>

This module allows plotting Skew-T/Log-P diagrams 
and hodographs from arrays of 
data with helper functions to plot directly from CM1 output files
in either Grads or HDF5 format and also from sounding data text
files in the format used for WRF and CM1 initialization.

This module also has code to produce rudimentary analysis of
the soundings data.  Currently this is limited to CAPE (surface 
and most unstable), CIN, wind shear, storm relative helicity, lifted index
and a rough estimate of storm motion based on the Bunkers (2000)
method.

.. figure:: _static/images/skewt.png
  :align: center

  Example Skew-T/Log-P with hodograph

Plotting
++++++++

High level plotting functions
-----------------------------

* :py:func:`plot_cm1h5` -- plots skewt from CM1 generated hdf5 files processed by ingest_cm1
* :py:func:`plot_cm1` -- plots skewt from CM1 generated GRaDS style output
* :py:func:`plot_wrf` -- plots skewt from WRF generated NetCDF output
* :py:func:`plot_sounding_data` -- plots skewt from CM1/WRF input sounding files
* :py:func:`plot_sounding_data_uwyo` -- plots skewt from uwyo sounding data (file or online)

These functions are called by command line scripts provided to make plotting from data files easy. 
You can invoke these through command line scripts as:

.. code-block:: bash

   # From tabular sounding data (e.g. WRF or CM1 intial sounding data)
   $ skewt tabular -f sounding.dat skewt.pdf

   # From GrADS stlye CM1 output
   $ skewt cm1 -p . -d cm1out -x 0 -y 0 skewt-cm1.pdf

   # From HDF5 CM1 output
   $ skewt cm1hdf -f model-data.h5 -x 0 -y 0 skewt.pdf

   # From WRF output 
   $ skewt wrf -f wrfout.nc --lat 30 --lon -80 -t 0 skewt.pdf

   # From University of Wyoming data file
   $ skewt uwyo -f uwyo-data.dat skewt.pdf

   # From University of Wyming website data (live)
   $ skewt uwyoweb --station 72251 skewt.pdf

* :py:func:`plot` -- generic high level plot function

.. code-block:: python

   import numpy as np
   import pymeteo.skewt as skewt

   # prepare 1D arrays height (z), pressure (p), potential temperature (th), 
   # water vapor mixing ratio (qv), winds (u and v) all of the same length.

   skewt.plot(None, z, th, p, qv, u, v, 'skewt.pdf')

Variables affecting the plot output
-----------------------------------

These variables affect the skewness of the plot and the bounds of the plot

* :py:data:`skew_angle` -- skewness
* :py:data:`Tmin`, :py:data:`Tmax` -- T dimension bounds (temperature, C)
* :py:data:`pbot`, :py:data:`ptop` -- p dimension bounds (pressure, Pa)

These variables affect the lines plotted on the skew-t

* :py:data:`isobars`
* :py:data:`isotherms`
* :py:data:`dry_adiabats`
* :py:data:`moist_adiabats`
* :py:data:`mixing_ratios`

These variables affect the plotting of the lines above

* :py:data:`lw_major`, :py:data:`lw_minor`
* :py:data:`lc_major`, :py:data:`lc_minor`

These variables affect the plotting style of skewt data

* :py:data:`linecolor_T`, :py:data:`linewidth_T`
* :py:data:`linecolor_Tve`, :py:data:`linewidth_Tve`, :py:data:`linestyle_Tve`
* :py:data:`linecolor_Td`, :py:data:`linewidth_Td`
* :py:data:`linecolor_Twb`, :py:data:`linewidth_Twb`
* :py:data:`linecolor_Parcel_T`, :py:data:`linewidth_Parcel_T`
* :py:data:`linecolor_Tvp`, :py:data:`linewidth_Tvp`, :py:data:`linestyle_Tvp`


Functions to draw isolines
--------------------------

If you are manually plotting skew-t data, these functions can be used to draw
various lines on your plot.  These are used by the high level plotting functions.

* :py:func:`draw_isotherms`
* :py:func:`draw_isobars`
* :py:func:`draw_dry_adiabat`
* :py:func:`draw_moist_adiabat`
* :py:func:`draw_water_mix_ratio`

Module reference
++++++++++++++++

"""
import os
import math
import pymeteo.thermo as met
import pymeteo.dynamics as dyn
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
import h5py
import pymeteo.cm1.read_grads as cm1
import pymeteo.interp
import pymeteo.constants as metconst
from netCDF4 import Dataset
import datetime as dt
import pymeteo.wrf as wrf
import datetime
import pymeteo.uwyo as uwyo

# This defines the skew-angle of the T axis
skew_angle = 37.5 
"""This defines the skewness of the T axis"""

# These define the domain plotted.  T values are at @1000 mb (e.g. unskewed)
Tmin = -40.0
"""Sets the left boundary of the plot. Temperature at 1000 mb (C)"""
Tmax = 40.0
"""Sets the right boundary of the plot. Temperature at 1000 mb (C)"""
pbot = 105000.0
"""Sets the bottom boundary of the plot. Pressure (Pa)"""
ptop = 10000.0
"""Sets the top boundary of the plot. Pressure (Pa)"""

## Values below used for plotting 
dp = 5000.0
"""The delta pressure used in calculating adiabats"""
ptickbot = 100000.0
"""Lowest elevated pressure level to be labelled in the plot (Pa)"""
pticktop = 10000.0
"""Highest elevated pressure level to be labelled in the plot (Pa)"""
tickdp = 10**4
"""The spacing between pressure labels in the plot (Pa)"""
plevs = np.arange(pbot,ptop-1,-dp)
"""List of pressure levels to do calculations on"""
fontscalefactor = 1

## Lines to plot
isotherms    = np.arange(-150,51,10)              # in degrees C
"""List of isotherms to plot on the skew-t.  In degrees C"""
isobars      = np.arange(ptickbot,ptop-1,-5000)   # in Pa
"""List of isobars to plot on the skew-t.  In Pascals"""
dry_adiabats = np.arange(-40,210,10)              # in degrees C
"""List of dry adiabats to plot on the skew-t.  In degrees C"""
moist_adiabats = np.concatenate((np.arange(-15.,10.1,5.),np.arange(12.,45.1,2.)))
"""List of moist adiabats to plot on the skew-t.  In degrees C"""
mixing_ratios = [0.2,0.4,0.8,1,2,3,4,6,8,10,14,18,24,32,40]
"""List of water vapor mixing ratio lines to plot. In g/kg"""

## Linewidths
lw_major = 0.6
"""Line width of 'major' lines.  E.g. Lines plotted at 10 C intervals or 50 mb intervals"""
lw_minor = 0.25
"""Line width of 'minor' lines.  E.g. Lines not plotted at the major intervals"""

## Linecolors
lc_major = 'grey'
"""Line color of 'major' lines.  E.g. Lines plotted at 10 C intervals or 50 mb intervals"""
lc_minor = 'lightgrey'
"""Line color of 'minor' lines.  E.g. Lines not plotted at the major intervals"""

## Skew-T line parameters
linecolor_T = 'black'
"""Line color of environmental temperature profile"""
linewidth_T = 1.5
"""Line width of environmental temperature profile"""

linecolor_Td = 'green'
"""Line color of environmental dew-point temperature profile"""
linewidth_Td = 1.5
"""Line width of environmental dew-point temperature profile"""

linecolor_Parcel_T = 'red'
"""Line color of lifted surface parcel temperature profile"""
linewidth_Parcel_T = 1.0
"""Line width of lifted surface parcel temperature profile"""

linecolor_Twb = 'blue'
"""Line color of environmental wet-bulb temperature profile"""
linewidth_Twb = 0.5
"""Line width of environmental wet-bulb temperature profile"""

linecolor_Tve = 'black'
"""Line color of environmental virtual temperature profile"""
linewidth_Tve = 0.7
"""Line width of environmental virtual temperature profile"""
linestyle_Tve = '--'
"""Line style of environmental virtual temperature profile"""

linecolor_Tvp = 'red'
"""Line color of lifted surface parcel virtual temperature profile"""
linewidth_Tvp = 0.7
"""Line width of lifted surface parcel virtual temperature profile"""
linestyle_Tvp = '--'
"""Line style of lifted surface parcel virtual temperature profile"""

#for plotted lines
pb_plot=105000
pt_plot=10000
pt_plot2=20000
dp_plot=1000
plevs_plot = np.arange(pb_plot,pt_plot-1,-dp_plot)
plevs_plot2 = np.arange(pb_plot,pt_plot2-1,-dp_plot)
plevs_std = [100000,85000,70000,50000,40000,30000,25000,20000,15000]

#TODO: enforce square w/ aspect ratio of plot
#Domain of the hodograph
umin = -22.5
umax = 27.5
vmin = -12.5
vmax = 27.5

##################################################################################
def plot_cm1h5(filename, xi, yi, output):
    """ Plots a skewt from an HDF5 file.

    :param filename: The name of the HDF5 file to open.
    :type filename: str
    :param xi: The X gridpoint of the skewt plot.
    :type xi: int
    :param yi: The Y gridpoint of the skewt plot.
    :type yi: int
    :param output: Output filename to save plot
    :type output: str

    To use this function the HDF5 file must have the following
    variables:

    - *xh* -- x scalar grid locations
    - *yh* -- y scalar grid locations
    - *z* -- z grid locations
    - *time* -- timestep of file data (scalar)
    - *th* -- potential temperature (K)
    - *thpert* -- potential temperature perturbation (K)
    - *prs* -- pressure (Pa)
    - *u* -- u wind speed (m/s)
    - *v* -- v wind speed (m/s)
    - *qv* -- water vapor mixing ratio (kg/kg)

    The names of these variables correspond to default naming by 
    CM1 using HDF5 output.

    """
    f = h5py.File(filename, 'r')
    z = f["/mesh/zh"][:]    # m 

    x = f["/mesh/xh"][xi]   # m
    y = f["/mesh/yh"][yi]   # m
    t = f["/time"][0]       # s
    th = f["/3d_s/thpert"][:,yi,xi] + f["/basestate/th0"][:] # K
    p = f["/3d_s/ppert"][:,yi,xi] + f["/basestate/pres0"][:] # Pa
    u = f["/3d_u/u"][:,yi,xi] # m/s
    v = f["/3d_v/v"][:,yi,xi] # m/s
    qv = f["/3d_s/qvpert"][:,yi,xi] + f["/basestate/qv0"][:] #kg/kg

    print(x,y,z[0],t,th[0],u[0],v[0],p[0],qv[0])
    plot_old(x,y,z,t,th,p,qv,u,v,filename, output)

##################################################################################
def plot_wrf(filename, lat, lon, time, output):
    """ Plots a skewt from an WRF NetCDF output file.

    :param filename: The name of the NetCDF file to open.
    :type filename: str
    :param lat: The latitude of the location of the skewt plot.
    :type lat: float
    :param lon: The longitude of the location of the skewt plot.
    :type lon: int
    :param output: Output filename to save plot
    :type output: str

    This function assumes the NetCDF file was produce by the WRF model.
    If another program produces the file but names variables in the same
    manner as the WRF model, this function should work for that dat as well.

    The variables using in the wrfout.nc data are:
    - Times ?
    - U(Time, bottom_top, south_north, west_east_stag)
    - V(Time, bottom_top, south_north_stag, west_east)
    - T(Time, bottom_top, south_north, west_east)   Theta perturbation
    - P(Time, bottom_top, south_north, west_east)   Pres perturbation
    - PB(Time, bottom_top, south_north, west_east)  pres base state
    - QVAPOR(Time, bottom_top, south_north, west_east)

    - PH(Time, bottom_top_stag, south_north, west_east)
    - PHB(Time, bottom_top_stag, south_north, west_east)
      Z = (((PH(k)+PH(k+1)) / 2) + ((PHB(k)+(PHB(k+1)) / 2) / 9.81

    - TH2(Time, south_north, west_east) -- 2m pot temp
    - PSFC(Time, south_north, west_east) -- surface pres
    - U10(Time, south_north, west_east) -- 10m U
    - V10(Time, south_north, west_east) -- 10m V
    - XTIME(Time) -- minutes since start of sim

    - XLAT, XLONG, XLAT_U/V, XLONG_U/V
    - Start_date

    Need time, theta, pressure
    """

    # Open NetCDF file
    f = Dataset(filename, 'r')

    wrf_lats = f.variables['XLAT'][time,:,:] # :,0
    wrf_lons = f.variables['XLONG'][time,:,:]# 0,:
    wrf_time = (f.variables['Times'][time]).tostring().decode('UTF-8')

    # refactor this into a wrf utility module

    map_proj = f.getncattr('MAP_PROJ')
    truelat1 = f.getncattr('TRUELAT1')
    truelat2 = f.getncattr('TRUELAT2')
    stand_lon= f.getncattr('STAND_LON')
    dx = f.getncattr('DX')
    dy = f.getncattr('DY')
    ref_lat = wrf_lats[0,0]
    ref_lon = wrf_lons[0,0]

    i,j = wrf.ll_to_ij(map_proj, truelat1, truelat2, stand_lon, dx, dy, ref_lat, ref_lon, lat, lon)

    ref_lat_u = f.variables['XLAT_U'][time,0,0]
    ref_lon_u = f.variables['XLONG_U'][time,0,0]
    ref_lat_v = f.variables['XLAT_V'][time,0,0]
    ref_lon_v = f.variables['XLONG_V'][time,0,0]

    i_u,j_u = wrf.ll_to_ij(map_proj, truelat1, truelat2, stand_lon, dx, dy, ref_lat_u, ref_lon_u, lat, lon)
    i_v,j_v = wrf.ll_to_ij(map_proj, truelat1, truelat2, stand_lon, dx, dy, ref_lat_v, ref_lon_v, lat, lon)

    # location
    N = 'N'
    if (lat < 0.):
        N = 'S'
    E = 'E'
    if (lon < 0.):
        E = 'W'
    x = "{0} {1}, {2} {3}".format(abs(lat), N, abs(lon), E)
    
    # pressure
    p_surface = f.variables['PSFC'][time,j,i]
    p = f.variables['P'][time,:,j,i] + f.variables['PB'][time,:,j,i]
    p = np.insert(p, 0, p_surface)

    # z heights
    ph = f.variables['PH'][time,:,j,i]
    phb = f.variables['PHB'][time,:,j,i]
    z_surface = ph[0]
    z = np.zeros(len(ph)-1, np.float32)
    for k in range(len(z)):
         z[k] = ((( ph[k]+ph[k+1] ) / 2.0) + ( phb[k] + phb[k+1]) / 2.0 ) / 9.81
    z = np.insert(z, 0, z_surface)

    # t
    t = wrf_time

    # theta
    th_surface = f.variables['TH2'][time,j,i]
    th = f.variables['T'][time,:,j,i] + 300.0
    th = np.insert(th, 0, th_surface)

    # u
    u_surface = f.variables['U10'][time,j,i]
    u = f.variables['U'][time,:,j_u,i_u]
    u = np.insert(u, 0, u_surface)

    # v
    v_surface = f.variables['V10'][time,j,i]
    v = f.variables['V'][time,:,j_v,i_v]
    v = np.insert(v, 0, v_surface)

    # qv
    qv_surface = f.variables['Q2'][time,j,i]
    qv = f.variables['QVAPOR'][time,:,j,i]
    qv = np.insert(qv, 0, qv_surface)

    title = os.path.basename(filename)
    print(x, z[0],t,th[0],u[0],v[0],p[0],qv[0])
    plot(x, z, th, p, qv, u, v, output, t, title)

    
##################################################################################
#
def plot_sounding_data(filename, output):
        """Plot SkewT from a WRF / CM1 compatible sounding data file
    
        :param filename: The name of the file to open.
        :type filename: str
        :param output: The name of the file to output plot
        :type output: str

        The datafile is the same format as used in sounding initalization
        files for WRF and CM1. 
        
        The format is:
        1 line header with surface pressure (mb), theta (K) and qv (g/kg)
        n number of lines with z (m), theta (K), qv (g/kg), u (m/s), v(m/s)

        """
        # load first line of file
        with open(filename, 'r') as f:
            surface = f.readline()

        p0, th0, qv0 = surface.split()

        # load rest of file
        _z, _th, _qv, _u, _v = np.loadtxt(filename, unpack=True, skiprows=1)

        # create arrays with one more z index for surface values
        nk = len(_z) + 1
        z = np.empty(nk, np.float32)
        th= np.empty(nk, np.float32)
        qv= np.empty(nk, np.float32)
        u = np.empty(nk, np.float32)
        v = np.empty(nk, np.float32)
        p = np.empty(nk, np.float32)

        # copy the arrays, leaving room at the surface
        z[1:nk] = _z
        th[1:nk] = _th
        qv[1:nk] = _qv / 1000.
        u[1:nk] = _u
        v[1:nk] = _v

        # assign surface values
        z[0] = 0.
        th[0] = float(th0)
        qv[0] = float(qv0) / 1000.
        u[0] = 1.75*u[1]-u[2]+0.25*u[3]
        v[0] = 1.75*v[1]-v[2]+0.25*v[3]
        p[0] = float(p0) * 100.

        # integrate pressure, assume hydrostatic
        # dp = -rho*g*dz
        for k in np.arange(1,nk):
            p[k] = p[k-1] * np.exp((metconst.g*(z[k-1]-z[k]))/(metconst.Rd*met.T((th[k]+th[k-1])/2.,p[k-1])))

        #for k in np.arange(nk):
        #    print(z[k], p[k], th[k], qv[k], u[k], v[k])
            
        plot(None, z, th, p, qv, u, v, output, title="input sounding")

def plot_sounding_data_uwyo(filename, output, stationID=0, date=None):
    """Plot SkewT from University of Wyoming sounding data

    :param filename: The name of the file containing sounding data
    :type filemane: str
    :param output: The name of the file to output plot
    :type output: str
    :param stationID: The station ID of a sounding station
    :type stationID: int
    :param date: Date and time of a sounding to request
    :type date: datetime

    If filename is not None then this function will plot data from that file.
    If filename is None and stationID is non-zero, then this will request a sounding
    datafile from http://weather.uwyo.edu for the specified date.  If date is None 
    then the requested sounding will be the most recent sounding taken at either
    12Z or 00Z. 

    """

    if (filename != None):
        title, p, z, qv, wind_dir, wind_speed, th = uwyo.fetch_from_file(filename)
    elif (stationID != 0):
        if date is None:
            date = datetime.datetime.utcnow()
        title, p, z, qv, wind_dir, wind_speed, th = uwyo.fetch_from_web(date, stationID)
    else:
        print("Neither input file or station ID was provided.  No output.")
        
    p, z, qv, u, v, th = uwyo.transform_and_check_data(p, z, qv, wind_dir, wind_speed, th)    
    plot(None, z, th, p, qv, u, v, output, title=title)
    
    
        
def plot_sounding_data_csv(filename, output):
        """Plot SkewT from a CSV sounding data file
    
        :param filename: The name of the file to open.
        :type filename: str
        :param output: The name of the file to output plot
        :type output: str

        The datafile format is CSV with the following columns: 

        - pressure (mb)
        - height (m)
        - temperature (C)
        - dew point (C)
        - wind direction (degrees)
        - wind speed (m/s)
        
        Missing values should be filled with the value -9999.00
        """
        p,z,T,Td,wdir,wspd = np.loadtxt(filename, delimiter=',',  unpack=True)
        # Pressure to Pa
        p = p * 100.

        # z to km
        #z = z / 1000.
        
        # interpolate missing wind

        nk = len(z)
        u = np.empty(nk, np.float32)
        v = np.empty(nk, np.float32)

        for k in range(nk):
           if wdir[k] == -9999. and wspd[k] == -9999.:
              u[k] = v[k] = -9999.
           else:
              u[k], v[k] = dyn.wind_deg_to_uv(wdir[k], wspd[k])

           #print('{0:5.2f} {1:5.2f} = {2:5.2f} {3:5.2f}'.format(wdir[k], wspd[k], u[k], v[k]))
           
        _z = np.empty(2,np.float32)
        _u = np.empty(2,np.float32)
        _v = np.empty(2,np.float32)
        print('INTERPOLATING')
        for k in range(nk):
           if wdir[k] == -9999. and wspd[k] == -9999.:
              kb = ke = k
              while kb >= 0 and wdir[kb] == -9999. and wspd[kb] == -9999.:
                 kb -= 1
              while ke <= nk-1 and wdir[ke] == -9999. and wspd[ke] == -9999.:
                 ke += 1

              # everything in bounds
              if kb >= 0 and ke <= nk-1:
                 _z[0] = z[kb]
                 _z[1] = z[ke]
                 _u[0] = u[kb]
                 _u[1] = u[ke]
                 _v[0] = v[kb]
                 _v[1] = v[ke]
                 
                 u[k] = pymeteo.interp.linear(_z, _u, z[k])
                 v[k] = pymeteo.interp.linear(_z, _v, z[k])

              elif kb < 0:
                 u[k] = u[ke]
                 v[k] = v[ke] 
              elif ke > nk-1:
                 u[k] = u[kb]
                 v[k] = v[kb] 

        for k in range(nk):
           # kt to m/s
           u[k] = u[k] * 0.5144444
           v[k] = v[k] * 0.5144444
        #   print('{0:5.2f} {1:5.2f} = {2:5.2f} {3:5.2f}'.format(wdir[k], wspd[k], u[k], v[k]))

        # calc theta
        th = np.empty(nk, np.float32)
        # calc qv
        qv = np.empty(nk, np.float32)
        for k in range(nk):
           th[k] = met.theta(T[k]+met.T00, p[k]) 
           w = met.es(Td[k]+met.T00) / met.es(T[k]+met.T00)
           pp = met.es(T[k]+met.T00) / p[k]
           qv[k] = 0.622 * pp * w
           #qv[k] = met.es(Td[k]+met.T00) / (met.Rv * (T[k]+met.T00))

        #print(z, th, p, qv, u, v)

        plot(None, z, th, p, qv, u, v, output, title='Sounding Data')

        
##################################################################################
# plot_cm1
#
# This plots a skewt at domain location xi,yi at time t=0 for a given CM1 dataset
# in grads format
#
def plot_cm1(path, filename, xi, yi,output):
  """Plot skewt from a Grads format CM1 output file

  :parameter _path: Path to CM1 dataset
  :type path: str
  :parameter _filename: Filename of dataset
  :type filename: str
  :parameter x1: X gridpoint to plot SkewT
  :parameter y1: Y gridpoint to plot SkewT
  :parameter output: Filename to save skewt plot

  This function plots a skewt from a CM1 output file.  
  This routine uses winds interpolated to the scalar
  gridpoints.  
  """
  f = cm1.CM1(path, filename)
  _z = f.dimZ[:] * 1000.   # km->m

  x = f.dimX[xi]
  y = f.dimY[yi]
  t = int(f.dimT[0])
        
  f.read3dMultStart(t)
  _th = f.read3dMult('th')
  _p = f.read3dMult('prs')
  _u = f.read3dMult('uinterp')
  _v = f.read3dMult('vinterp')
  _qv = f.read3dMult('qv')
  f.read3dMultStop()
  nk = len(_p[xi,yi,:])
  th = np.empty(nk+1, np.float32)
  p  = np.empty(nk+1, np.float32)
  u  = np.empty(nk+1, np.float32)
  v  = np.empty(nk+1, np.float32)
  qv = np.empty(nk+1, np.float32)
  z  = np.empty(nk+1, np.float32)
  th[1:] = _th[xi,yi,:]
  p[1:] = _p[xi,yi,:]
  u[1:] = _u[xi,yi,:]
  v[1:] = _v[xi,yi,:]
  qv[1:] = _qv[xi,yi,:]
  z[1:] = _z[:]

  # surface values
  p[0] = 100000.
  th[0]= 300.
  qv[0]= 0.014
  u[0] = u[1]
  v[0] = v[1]
  z[0] = 0.

  plot_old(x,y,z,t,th,p,qv,u,v,filename, output)

##################################################################################
# plot
#
# This is the main skewT plotting function for a single output page containing
# A skewt, hodograph and an information block (currently disabled).
#
#TODO: turn x,y into "location text" at end of arg list default None
#TODO: put title at end default None.
#TODO: pass both to plot_the_rest
def plot_old(x, y, z, time, th, p, qv, u, v, title, output):
  plot("{0} km, {1} km".format(x,y), z, th, p, qv, u, v, output, time, title) 

def plot(loc, z, th, p, qv, u, v, output, time = None, title = None):
  """Plots Skew-T/Log-P diagrapms with hodograph

  The helper functions above facilitate loading data from
  various formats and then call this function.  If you have
  data in another format or arrays of data in python already,
  then this is the function you want to use.

  :parameter loc: Location string
  :parameter z: z grid mesh (1D)
  :parameter time: Time string
  :parameter th: Potential temperature at z points
  :parameter p: Pressure at z points
  :parameter qv: Water vapor mixing ratio at z points
  :parameter u: u winds at z points
  :parameter v: v winds at z points
  :parameter title: Title for plot
  :parameter output: Filename to save plot to

  """
  fig = plt.figure(1, figsize=(10, 8), dpi=300, edgecolor='k')
  
  # sounding
  ax1 = plt.subplot(121)
  plot_sounding_axes(ax1)
  plot_sounding(ax1, z, th, p, qv, None, None)
  # hodograph
  ax2 = plt.subplot(222)
  plot_hodo_axes(ax2)
  plot_hodograph(ax2, z, u, v)
  # datablock
  ax3 = fig.add_subplot(224)
  try:
    plot_datablock(ax3, loc, z, time, th, p, qv, u, v, title)
  except:
      print("Error calcualting sounding stats, datablock omitted");
    
  # wind barbs
  ax4 = fig.add_subplot(132)
  plot_wind_axes(ax4)
  plot_wind_barbs(ax4,z,p,u,v)
  # legend
  ax5 = fig.add_subplot(4,4,15)
  plot_legend(ax5)

  
  # Adjust plot margins.
  plt.subplots_adjust(left=0.03, bottom=0.03, right=0.97, top=0.97, wspace=0.12, hspace=0.12)
  plt.savefig(output, dpi=300,bbox_inches=0)
  plt.close()

def plot_sounding_axes(axes):
  """Plots Skew-T/Log-P axes

  This will plot isotherms, isobars, dry and moist adiabats, 
  lines of constant water vapor mixing ratio, labels and 
  setup the y axes to be reversed.

  :paramter axes: The axes to draw on
  """
  draw_isotherms(axes)
  draw_isobars(axes)
  draw_dry_adiabat(axes)
  draw_moist_adiabat(axes)
  draw_water_mix_ratio(axes)
  remove_tick_labels(axes)
  axes.axis([Tmin, Tmax, pbot, ptop])
  axes.set_ylim(axes.get_ylim()[::1])

def plot_wind_axes(axes):
  # plot wind barbs
  # TODO: also do storm-relative winds
  draw_wind_line(axes)
  axes.set_axis_off()
  axes.axis([-1,1,pbot,ptop])

def plot_hodo_axes(axes):
  """Plot hodograph axes

  This will plot range arcs and labels for a hodograph plot
  """
  bounds = [-25,25,-25,25]
  axes.axis('equal')
  draw_hodograph(axes, bounds)
  remove_tick_labels(axes)
  axes.axis(bounds)


def plot_legend(axes):
  """Plot skew-t legend"""

  tT = r'Temperature'
  lT = Line2D(range(10), range(10), linestyle='-', marker='', linewidth=linewidth_T, color=linecolor_T)

  tTd = r'Dew-point Temperature'
  lTd = Line2D(range(10), range(10), linestyle='-', marker='', linewidth=linewidth_Td, color=linecolor_Td)

  tPT = r'Lifted Surface Parcel Temperature'
  lPT = Line2D(range(10), range(10), linestyle='-', marker='', linewidth=linewidth_Parcel_T,
               color=linecolor_Parcel_T)

  tTwb = r'Wet-bulb Temperature'
  lTwb = Line2D(range(10), range(10), linestyle='-', marker='', linewidth=linewidth_Twb, color=linecolor_Twb)

  tTve = r'Virtual Temperature'
  lTve = Line2D(range(10), range(10), linestyle=linestyle_Tve, marker='', linewidth=linewidth_Tve,
                color=linecolor_Tve)

  tTvp = r'Lifted Surface Parcel Virtual Temperature'
  lTvp = Line2D(range(10), range(10), linestyle=linestyle_Tvp, marker='', linewidth=linewidth_Tvp,
                color=linecolor_Tvp)

  plt.legend((lT, lTve, lTd, lTwb, lPT, lTvp,),(tT, tTve, tTd, tTwb, tPT, tTvp,),
             loc=(0.125,0), fontsize=6, handlelength=10)
  # loc =, frameon=, fontsize=
  axes.set_axis_off()
  
  
def plot_wind(axes, z, p, u, v, x=0):  
  for i in np.arange(0,len(z),1):
    if (p[i] > pt_plot):
      plt.barbs(x,p[i],u[i],v[i], length=5, linewidth=.5)

  
def plot_sounding(axes, z, th, p, qv, u = None, v = None):
  """Plot sounding data

  This plots temperature, dewpoint and wind data on a Skew-T/Log-P plot.
  This will also plot derived values such as wetbulb temperature and
  label the surface based LCL, LFC and EL.

  :parameter z: height values (1D array)
  :parameter th: potential temperature at z heights (1D array)
  :parameter p: pressure at z heights (1D array)
  :parameter qv: water vapor mixing ratio at z heights (1D array)
  :parameter u: U component of wind at z heights (1D array)
  :parameter v: V component of wind at z heights (1D array)
  :paramter axes: The axes instance to draw on
  """
  # calculate Temperature and dewpoint
  T = met.T(th,p) - met.T00                          # T (C)
  Td = met.Td(p, qv) - met.T00                       # Td (C)

  # calculate wetbulb temperature
  Twb = np.empty(len(z), np.float32)                  # Twb (C)
  for zlvl in range(len(z)):
    Twb[zlvl] = met.Twb(z, p, th, qv, z[zlvl])

  # Get surface parcel CAPE and temperature / height profiles
  pcl = met.CAPE(z, p, T+met.T00, qv, 1)        # CAPE
  T_parcel = pcl['t_p'] - met.T00                      # parcel T (C)
  T_vparcel = pcl['tv_p'] - met.T00                     # parcel Tv (C)
  T_venv = met.T(pcl['thv_env'], pcl['pp']) - met.T00  # Env Tv (C)

  # plot Temperature, dewpoint, wetbulb and lifted surface parcel profiles on skew axes
  axes.semilogy(T + skew(p), p, base=math.e, color=linecolor_T , linewidth = linewidth_T)
  axes.semilogy(Td + skew(p), p, base=math.e, color=linecolor_Td, linewidth = linewidth_Td)
  axes.semilogy(T_parcel + skew(pcl['pp']), pcl['pp'], base=math.e,
                color=linecolor_Parcel_T, linewidth=linewidth_Parcel_T)
  axes.semilogy(Twb + skew(p), p, base=math.e, color=linecolor_Twb, linewidth=linewidth_Twb)

  # plot virtual temperature of environment and lifted parcel
  axes.semilogy(T_venv + skew(pcl['pp']), pcl['pp'], base=math.e, color=linecolor_Tve,
                linewidth=linewidth_Tve, linestyle=linestyle_Tve)
  axes.semilogy(T_vparcel + skew(pcl['pp']), pcl['pp'], base=math.e, color=linecolor_Tvp,
                linewidth=linewidth_Tvp, linestyle=linestyle_Tvp)

  # Add labels for levels based on surface parcel
  #debug print(pcl['lfcprs'], pcl['lclprs'], pcl['elprs'], pcl['ptops'])
  if (pcl['lfcprs'] > 0):
    label_m(Tmax-.5, pcl['lfcprs'], '--LFC', axes)
  if (pcl['lclprs'] > 0):
    label_m(Tmax-.5, pcl['lclprs'], '--LCL', axes)
  if (pcl['elprs'] > 0):
    label_m(Tmax-.5, pcl['elprs'], '--EL', axes)
  if (pcl['ptops'] > 0):
    label_m(Tmax-.5, pcl['ptops'], '--TOPS', axes)

  # plot labels for std heights
  for plvl in plevs_std:
    zlvl = pymeteo.interp.interp_height(z,p,plvl)
    label_m(Tmin-.5,plvl, str(int(zlvl)), axes)

  # plot wind barbs on left side of plot.  move this?  right side?
  if (u is not None and v is not None):
      #draw_wind_line(axes)
      for i in np.arange(0,len(z),2):
          if (p[i] > pt_plot):
              plt.barbs(Tmin+4,p[i],u[i],v[i], length=5, linewidth=.5)

def plot_wind_barbs(axes, z, p, u, v):
    for i in np.arange(0,len(z)):
        if (p[i] > pt_plot):
            plt.barbs(0,p[i],u[i],v[i], length=5, linewidth=.5)

              
def plot_hodograph(axes, z, u, v):
  """Plot Hodograph data

  This plots u and v winds vs height on a hodograph.

  :parameter z: height values (1D array)
  :parameter u: U component of wind at z heights (1D array)
  :parameter v: V component of wind at z heights (1D array)
  :paramter axes: The axes instance to draw on
  """
  
  # plot hodograph
  z6km = 0
  try:
    while z[z6km] <= 12000:
      z6km += 1
  except IndexError:
    z6km = len(z)-1
  axes.plot(u[0:z6km],v[0:z6km], color='black', linewidth=1.5)

  for zlvl in np.arange(0,7000,1000):
    ulvl = pymeteo.interp.linear(z,u,zlvl)
    vlvl = pymeteo.interp.linear(z,v,zlvl)
    #print('calculating winds at height {0} = ({1},{2})'.format(zlvl,ulvl,vlvl))
    label_h2(ulvl+1,vlvl-1,str(zlvl/1000), 'black', 0, axes)
    axes.plot(ulvl,vlvl, color='black', markersize=5, marker='.')

  #TODO: fix this
    try:
      ucb = dyn.storm_motion_bunkers(u,v,z)
      axes.plot(ucb[0],ucb[1],markersize=4,color='black',marker='x')
      axes.plot(ucb[2],ucb[3],markersize=4,color='black',marker='x')
    except:
        print("Error calculating sounding stats, storm motion marker not plotted");
      
def calc_sounding_stats(_z, _th, _p, _qv):
  T = met.T(_th,_p)                        # T (K)
  pcl = met.CAPE(_z, _p, T, _qv, 1)        # CAPE
  mupcl = met.CAPE(_z, _p, T, _qv, 2)      # MUCAPE
  mlpcl = met.CAPE(_z, _p, T, _qv, 3)      # MLCAPE

  return (pcl,mupcl,mlpcl)

def calc_hodograph_stats(_z, _u, _v):

  try:
      ucb = dyn.storm_motion_bunkers(_u,_v,_z)

# SRH
      srh01 = dyn.srh(_u, _v, _z, 0., 1000., ucb[0], ucb[1])
      srh03 = dyn.srh(_u, _v, _z, 0., 3000., ucb[0], ucb[1])

  except:
      print("Error calculating storm motion")
      ucb = [0.,0.,0.,0.]
      srh01 = 0.
      srh03 = 0.
      
  try:
    erh01 = dyn.srh(_u, _v, _z, 0., 1000., 0., 0.)
    erh03 = dyn.srh(_u, _v, _z, 0., 3000., 0., 0.)
  
  except:
    print("Error calculating erh")
    erh01 = 0.
    erh03 = 0.

 
  shear01 = dyn.uv_to_deg(*dyn.shear(_u, _v, _z, 0., 1000.))
  shear03 = dyn.uv_to_deg(*dyn.shear(_u, _v, _z, 0., 3000.))
  shear06 = dyn.uv_to_deg(*dyn.shear(_u, _v, _z, 0., 6000.))
  shear12 = dyn.uv_to_deg(*dyn.shear(_u, _v, _z, 1000., 2000.))
  shear23 = dyn.uv_to_deg(*dyn.shear(_u, _v, _z, 2000., 3000.))
  shear34 = dyn.uv_to_deg(*dyn.shear(_u, _v, _z, 3000., 4000.))
  shear45 = dyn.uv_to_deg(*dyn.shear(_u, _v, _z, 4000., 5000.))
  shear56 = dyn.uv_to_deg(*dyn.shear(_u, _v, _z, 5000., 6000.))

  dict = { 'bunkers' : ucb,
           'srh01'   : srh01,
           'srh03'   : srh03,
           'erh01'   : erh01,
           'erh03'   : erh03,
           's01'     : shear01,
           's03'     : shear03,
           's06'     : shear06,
           's12'     : shear12,
           's23'     : shear23,
           's34'     : shear34,
           's45'     : shear45,
           's56'     : shear56
         }

  return dict


def plot_datablock(ax4, _x,_z,_t,_th,_p,_qv,_u,_v, _title):
  pcl, mupcl, mlpcl = calc_sounding_stats(_z, _th, _p, _qv)
  shear = calc_hodograph_stats(_z, _u, _v)

  brn = dyn.brn(_u, _v, _z, pcl['cape'])

  # draw datablock
  ax4.set_axis_off()
  plt.axis([-1,1,-1,1])
  plt.text(0,1,_title, verticalalignment='top', horizontalalignment='center', weight='bold', fontsize=10)
  line = ""
  if (_x != None):
    line += "Sounding at location " + str(_x) + "."
  if (_t != None):
    line += " Time = " + str(_t) + "."
  line += " " + str(len(_z)) + ' vertical levels'
  plt.text(0,.85, line, verticalalignment='center', horizontalalignment='center', fontsize=5)

  cth,cr = dyn.uv_to_deg(shear['bunkers'][0],shear['bunkers'][1])
  line = 'Estimated storm motion (supercell right mover) -> {0:3d}$^\circ$ {1:3.1f} m/s'.format(int(cth),cr)
  plt.text(-1,.75, line, verticalalignment='center', horizontalalignment='left', fontsize=5)

  print_parcel_info('Surface Parcel', pcl, -1., .65)
  print_parcel_info('Most Unstable Parcel', mupcl, -0.3, .65)
  print_parcel_info('500 m Mixed Layer Parcel', mlpcl, 0.4, .65)

	# LCL, CCL, EL, convective temp?
	# other data?
	
  x = 0.4
  y = 0
  line = 'Hodograph'
  plt.text(x,y, line, verticalalignment='center', horizontalalignment='left', fontsize=5)
  x += 0.02
  y -= 0.065
  line = '0-1 km shear {0:3d}$^\circ$ {1:3.1f} m/s'.format(int(shear['s01'][0]),shear['s01'][1])
  plt.text(x,y, line, verticalalignment='center', horizontalalignment='left', fontsize=5)
  y -= 0.05
  line = '0-3 km shear {0:3d}$^\circ$ {1:3.1f} m/s'.format(int(shear['s03'][0]),shear['s03'][1])
  plt.text(x,y, line, verticalalignment='center', horizontalalignment='left', fontsize=5)
  y -= 0.05
  line = '0-6 km shear {0:3d}$^\circ$ {1:3.1f} m/s'.format(int(shear['s06'][0]),shear['s06'][1])
  plt.text(x,y, line, verticalalignment='center', horizontalalignment='left', fontsize=5)
  y -= 0.05
  line = 'SRH 0-1 : {0:d} m2/s2'.format(int(shear['srh01']))
  plt.text(x,y, line, verticalalignment='center', horizontalalignment='left', fontsize=5)
  y -= 0.05
  line = 'SRH 0-3 : {0:d} m2/s2'.format(int(shear['srh03']))
  plt.text(x,y, line, verticalalignment='center', horizontalalignment='left', fontsize=5)
  y -= 0.05
  line = 'ERH 0-1 : {0:d} m2/s2'.format(int(shear['erh01']))
  plt.text(x,y, line, verticalalignment='center', horizontalalignment='left', fontsize=5)
  y -= 0.05
  line = 'ERH 0-3 : {0:d} m2/s2'.format(int(shear['erh03']))
  plt.text(x,y, line, verticalalignment='center', horizontalalignment='left', fontsize=5)
  y -= 0.05
  line = 'BRN : {0:d}'.format(int(brn))
  plt.text(x,y, line, verticalalignment='center', horizontalalignment='left', fontsize=5)
  y -= 0.05
  line = '0-1 km shear {0:3d}$^\circ$ {1:3.1f} m/s'.format(int(shear['s01'][0]),shear['s01'][1])
  plt.text(x,y, line, verticalalignment='center', horizontalalignment='left', fontsize=5)
  y -= 0.05
  line = '1-2 km shear {0:3d}$^\circ$ {1:3.1f} m/s'.format(int(shear['s12'][0]),shear['s12'][1])
  plt.text(x,y, line, verticalalignment='center', horizontalalignment='left', fontsize=5)
  y -= 0.05
  line = '2-3 km shear {0:3d}$^\circ$ {1:3.1f} m/s'.format(int(shear['s23'][0]),shear['s23'][1])
  plt.text(x,y, line, verticalalignment='center', horizontalalignment='left', fontsize=5)
  y -= 0.05
  line = '3-4 km shear {0:3d}$^\circ$ {1:3.1f} m/s'.format(int(shear['s34'][0]),shear['s34'][1])
  plt.text(x,y, line, verticalalignment='center', horizontalalignment='left', fontsize=5)
  y -= 0.05
  line = '4-5 km shear {0:3d}$^\circ$ {1:3.1f} m/s'.format(int(shear['s45'][0]),shear['s45'][1])
  plt.text(x,y, line, verticalalignment='center', horizontalalignment='left', fontsize=5)
  y -= 0.05
  line = '5-6 km shear {0:3d}$^\circ$ {1:3.1f} m/s'.format(int(shear['s56'][0]),shear['s56'][1])
  plt.text(x,y, line, verticalalignment='center', horizontalalignment='left', fontsize=5)

def print_3col(name, value, unit, x, y):
   plt.text(x,y, name, verticalalignment='center', horizontalalignment='left', fontsize=5)
   plt.text(x+.35,y, value, verticalalignment='center', horizontalalignment='right', fontsize=5)
   plt.text(x+.4,y, unit, verticalalignment='center', horizontalalignment='left', fontsize=5)

def print_parcel_info(title, pcl, x, y):
  plt.text(x,y, title, verticalalignment='center', horizontalalignment='left', fontsize=5)
  y -= 0.065
  x += 0.02
  print_3col('CAPE', '{0}'.format(int(pcl['cape'])), 'J kg$^{-1}$', x, y)
  y -= 0.05
  print_3col('CIN', '{0}'.format(int(pcl['cin'])), 'J kg$^{-1}$', x, y)
  y -= 0.05
  print_3col('TOPS', '{0:4.2f}'.format(float(pcl['ztops'])), 'km', x, y)
  y -= 0.05
  print_3col(r'$\theta_e$', '{0:4.1f}'.format(float(pcl['theta_e'])), 'K', x, y)
  y -= 0.05
  print_3col('LI$_{MAX}$', '{0:3.1f}'.format(float(pcl['max_li'])), 'C', x, y)
  y -= 0.05
  print_3col('LI$_{500}$', '{0:3.1f}'.format(float(pcl['li500'])), 'C', x, y)
  y -= 0.05
  print_3col('LI$_{300}$', '{0:3.1f}'.format(float(pcl['li300'])), 'C', x, y)
  y -= 0.05
  print_3col('Parcel', '{0}'.format(int(pcl['prs']/100.)), 'mb', x, y)


def remove_tick_labels(axes):
  axes.tick_params(axis='x', top='off', bottom='off', which='both')#, labelsize=0)
  axes.tick_params(axis='y', left='off', right='off', which='both')#, labelsize=0)
  for xlabel_i in axes.get_xticklabels():
    xlabel_i.set_visible(False)
    xlabel_i.set_fontsize(0.0)
  for xlabel_i in axes.get_yticklabels():
    xlabel_i.set_fontsize(0.0)
    xlabel_i.set_visible(False)

def set_fontscalefactor(newfactor):
  fontscalefactor = newfactor

def label(x, y, s, c, r, axes):
	axes.text(x,y*100,s, verticalalignment='center', horizontalalignment='center', weight='bold', fontsize=5*fontscalefactor, color=c, rotation=r)

def label_h(x, y, s, c, r, axes):
	axes.text(x,y,s, verticalalignment='center', horizontalalignment='center', weight='normal', fontsize=7*fontscalefactor, color=c, rotation=r)

def label_h2(x, y, s, c, r, axes):
	axes.text(x,y,s, verticalalignment='center', horizontalalignment='center', weight='bold', fontsize=10*fontscalefactor, color=c, rotation=r)

def label_m(x, y, s, axes):
	axes.text(x,y,s, verticalalignment='center', horizontalalignment='right', fontsize=4*fontscalefactor)

#draw hodograph u-v space
def draw_hodograph(axes, bounds):
  xmin, xmax, ymin, ymax = bounds
  gmax = max(xmax,ymax)
  gmin = min(xmin,ymin)
	# draw u-v grid
  axes.plot([gmin,gmax],[0,0], color='black', linewidth=.5)
  axes.plot([0,0],[gmin,gmax], color='black', linewidth=.5)
  for u in np.arange(xmin+1,xmax):
    if (u%5==0) and (u != 0):
      label_h(u,-1,str(u),'black',0,axes)
  for v in np.arange(ymin+1,ymax):
    if (v%5==0) and (v != 0):
      label_h(-1,v,str(v),'black',0,axes)
	# draw speed rings
  for u in np.arange(5,math.sqrt(xmax**2+ymax**2)+1,5):
    axes.plot( *hodograph_circle(u), color='grey', linewidth=.3)

# helper function for circle points
def hodograph_circle(r):
	phi = np.arange(0., 2*math.pi, 0.01)
	return r*np.cos(phi), r*np.sin(phi)

# draws a vertical line axis upon which to draw wind barbs
def draw_wind_line(axes):
	wind_line = []
	for p in plevs_plot:
		wind_line.append(0)
	axes.semilogy(wind_line, plevs_plot, color='black', linewidth=.5)
	# plot circles at certain levels?
	for p in plevs_std:
		axes.semilogy(0,p, color='black', markersize=3, marker='.')

# Puts the skew in skew-T
def skew(p):
    """Puts the skew in skew-T

    :parameter p: pressure level of the point.

    This calculates the skew of the T axis for a point to plot.  
    This assumes a logarithmic y axes and uses the variable
    :py:data:skew_angle to determine the skew.  This is the 
    magic that turns a cartesian plot into a Skew-T/Log-p plot.
    """
    return skew_angle * np.log(met.p00/p)

# Draw isotherms on skew-T / log p axes
def draw_isotherms(axes):
    """Plot isotherms on axes

    :parameter axes: The axes to draw on
    :type axes: :py:class:`matplotlib.axes`

    This function draws isotherms every 10 C.
    """
    for T in isotherms:
        if (T % 10 == 0):
           axes.semilogy(T + skew(plevs_plot), plevs_plot, base=math.e, color = lc_major, linewidth= lw_major)
        else:
           axes.semilogy(T + skew(plevs_plot), plevs_plot, base=math.e, color = lc_minor, linewidth= lw_minor)
    for T in np.arange(-40, 40, 10):
        label(T+skew(87500),875, str(T), 'red', 90.-skew_angle, axes)
    for T in np.arange(-100, -20, 10):
        label(T+skew(17500),175, str(T), 'red', 90.-skew_angle, axes)

def draw_isobars(axes):
    """Plot isobars on axes

    :parameter axes: The axes to draw on
    :type axes: :py:class:`matplotlib.axes`

    This function draws isobars at intervals of 2*dp.
    """
    for i in isobars:
        if (i % 5000 == 0):
            axes.plot([Tmin, Tmax], [i,i], color = lc_major, linewidth = lw_major)
        else:
            axes.plot([Tmin, Tmax], [i,i], color = lc_minor, linewidth = lw_minor)
    for i in np.arange(1000,100,-50):
        label(-10-((1000-i)*.025),i,str(i),'black',0, axes)

def draw_dry_adiabat(axes):
    """Plot dry adiabats on axes

    :parameter axes: The axes to draw on
    :type axes: :py:class:`matplotlib.axes`

    This function calculates dry adiabats
    and plots these lines.  Adiabats are calculated 
    every 10 K
    """
    for T in dry_adiabats:
        dry_adiabat = met.T(T+met.T00,plevs_plot) - met.T00 + skew(plevs_plot)
        if (T % 10 == 0):
            axes.semilogy(dry_adiabat, plevs_plot, base=math.e, color = lc_major, linewidth = lw_major)
        else:
            axes.semilogy(dry_adiabat, plevs_plot, base=math.e, color = lc_minor, linewidth = lw_minor)
            
    for T in np.arange(-20, 150, 10):
        p = (600. - 3.5*T)*100.
        x = met.T(T+met.T00,p) -met.T00 + skew(p)
        x1 = met.T(T+met.T00,p+.5*dp_plot) -met.T00 + skew(p+.5*dp_plot)
        x2 = met.T(T+met.T00,p-.5*dp_plot) -met.T00 + skew(p-.5*dp_plot)
        dx = x2-x1
        theta = math.atan2(-dp_plot,-dx) * 180/math.pi +37
        label(x,p/100,str(T),'black',theta, axes)


def draw_moist_adiabat(axes):
    """Plot moist adiabats on axes

    :parameter axes: The axes to draw on
    :type axes: :py:class:`matplotlib.axes`

    This function calculates moist adiabats
    and plots these lines.  Adiabats are calculated for
    values of T at 1000mb from -15 to 45 C every 5 C between
    -15 and 10 C and every 2.5 C between 12.5 and 45 C.
    """
    ps_blo = [p for p in plevs_plot if p > 100000]
    ps_blo.reverse()
    ps = [p for p in plevs_plot2 if p < 100000]
    for T in moist_adiabats:
        T_1000 = T = T + met.T00
        moist_adiabat = []
        # work backwards from 1000mb
        for p in ps_blo:
            T += met.dTdp_moist(T,p) * dp_plot
            moist_adiabat.append(T - met.T00 + skew(p))
        #reverse list order
        moist_adiabat.reverse()
        # insert 1000mb point
        T = T_1000
        moist_adiabat.append(T - met.T00)
        # work forwards from 1000mb
        for p in ps:
            T -= met.dTdp_moist(T,p) * dp_plot
            moist_adiabat.append(T - met.T00 + skew(p))
            # draw labels
            if (p == 22000):
                if (T_1000 >= met.T00 and T_1000 <= 30+met.T00):
                    label(T-met.T00+skew(p),p/100,str(int(T_1000-met.T00)),'green', 0, axes)
        if (int(T_1000 - met.T00) % 5 == 0):            
            axes.semilogy(moist_adiabat, plevs_plot2, base=math.e, color = lc_major, linewidth = lw_major)
        else:
            axes.semilogy(moist_adiabat, plevs_plot2, base=math.e, color = lc_minor, linewidth = lw_minor)


def draw_water_mix_ratio(axes):
    """Plot lines of constant water vapor mixing ratio on axes

    :parameter axes: The axes to draw on
    :type axes: :py:class:`matplotlib.axes`

    This function calculates isolines of constant water vapor
    mixing ratio and plots these lines.  Values of w calculated
    are given by the list variable w.
    """
    #TODO: put w and the top plevel for plotting somewhere configurable
    ps = [p for p in plevs if p>=20000 and p<=105000]
    for W in mixing_ratios:
        water_mix = []
        for p in ps:
            T = TMR(W,p/100.) 
            water_mix.append(T + skew(p))
        axes.semilogy(water_mix, ps, base=math.e, color = 'grey', linestyle = '--', linewidth = .5)

        # Label the isoline
        T = TMR(W,1075.)
        label(T+skew(107500.), 1075, str(W), 'black', -15, axes)

def label_std_heights(axes):
   """Plot heights of standard pressure levels 

   :paramter axes: The axes to draw on
   :type axes: :py:class:`matplotlib.axes`
   """
   xpos = Tmin+1.5
   std_heights = [(1000,111),(925,2512),(850,1457),(700,3012),(500,5574),(400,7185),
                 (300,9164),(250,10363),(200,11784),(150,13608)]#,(100,16180)]
   for p,z in std_heights:
      label(xpos, p, str(z), 'black', 0, axes)

def TMR(W, p):
  # Computes temperature on mixing ratio w at pressure p.
  # TMR in C, w in g/kg dry air, p in millibars.
  # TODO: change this to something else?
  x = np.log10(W * p / (622 + W))
  TMR = 10 ** (0.0498646455 * x + 2.4082965) - 280.23475 + 38.9114 * ((10 ** (0.0915 * x) - 1.2035) ** 2)
  return TMR





import sys
import inspect
from PyQt5 import QtWidgets, QtGui, QtCore
from PyQt5.QtCore import pyqtSignal, pyqtSlot

import pymeteo.cm1.soundings
import pymeteo.cm1.hodographs

from pymeteo import skewt
import pymeteo.cm1.PlotWidget as PW
import pymeteo.cm1.StatWidget as SW


class MainWindow(QtWidgets.QWidget):

   def __init__(self):
      super(MainWindow,self).__init__()
      self.initUI()

   def initUI(self):
      
      self.splitter_2 = QtWidgets.QSplitter(QtCore.Qt.Horizontal)

      self.LVert_widget = QtWidgets.QWidget(self.splitter_2)
      self.verticalLayout = QtWidgets.QVBoxLayout(self.LVert_widget)
      #self.verticalLayout.setMargin(0)
      self.horizontalLayout = QtWidgets.QHBoxLayout()
      self.label = QtWidgets.QLabel('Output file:',self.LVert_widget)
      self.horizontalLayout.addWidget(self.label)
      self.lineEdit = QtWidgets.QLineEdit(self.LVert_widget)
      self.horizontalLayout.addWidget(self.lineEdit)
      self.pushButton = QtWidgets.QPushButton("Export", self.LVert_widget)
      self.horizontalLayout.addWidget(self.pushButton)
      self.verticalLayout.addLayout(self.horizontalLayout)
      self.line = QtWidgets.QFrame(self.LVert_widget)
      self.line.setFrameShape(QtWidgets.QFrame.HLine)
      self.line.setFrameShadow(QtWidgets.QFrame.Sunken)
      self.verticalLayout.addWidget(self.line)
      self.SoundingTabs = QtWidgets.QTabWidget(self.LVert_widget)
      self.verticalLayout.addWidget(self.SoundingTabs)
      self.line_2 = QtWidgets.QFrame(self.LVert_widget)
      self.line_2.setFrameShape(QtWidgets.QFrame.HLine)
      self.line_2.setFrameShadow(QtWidgets.QFrame.Sunken)
      self.verticalLayout.addWidget(self.line_2)
      self.WindTabs = QtWidgets.QTabWidget(self.LVert_widget)
      self.verticalLayout.addWidget(self.WindTabs)
      self.drawButton = QtWidgets.QPushButton("Update Plot", self.LVert_widget)
      self.quitButton = QtWidgets.QPushButton("Quit", self.LVert_widget)
      self.verticalLayout.addWidget(self.drawButton)
      self.verticalLayout.addWidget(self.quitButton)
      self.splitter = QtWidgets.QSplitter(self.splitter_2)
      self.splitter.setOrientation(QtCore.Qt.Vertical)
      self.Hodograph = PW.PlotWidget(self.splitter)
      self.SoundingInfo = SW.StatWidget(self.splitter)
      self.Sounding = PW.PlotWidget(self.splitter_2)

      # Load all of the sounding modules into tabs
      for name,obj in inspect.getmembers(pymeteo.cm1.soundings):
        if inspect.ismodule(obj):
          cobj = getattr(obj,name)
          if inspect.isclass(cobj):
             instance = cobj()
             self.SoundingTabs.addTab(instance, name)
      
      # Load all of the hodograph modules into tabs
      for name,obj in inspect.getmembers(pymeteo.cm1.hodographs):
        if inspect.ismodule(obj):
          cobj = getattr(obj,name)
          if inspect.isclass(cobj):
             instance = cobj()
             self.WindTabs.addTab(instance, name)

      self.Sounding.SStats.connect(self.SoundingInfo.SStat_values)

      self.pushButton.clicked.connect(self.export)
      self.exportSounding.connect(self.SoundingInfo.output_sounding)

      self.drawButton.clicked.connect(self.update_plot)
      self.quitButton.clicked.connect(self.close)

      self.hbox = QtWidgets.QHBoxLayout(self)
      self.hbox.addWidget(self.splitter_2)
      self.setLayout(self.hbox)

      self.resize(1200,768)

      #TODO: size widgets initially. This doesnt work
      self.LVert_widget.resize(250,500)
      self.Hodograph.resize(250,250)
      self.SoundingInfo.resize(250,250)
      self.Sounding.resize(250,500)


      self.center()
      self.setWindowTitle('CM1 / WRF sounding initialization tool')

      self.show()
      self.Sounding.plot_sounding_axes()
      self.Hodograph.plot_hodograph_axes()
      self.SoundingInfo.calcStats()

      self.SoundingInfo.statsdone.connect(self.enableButton)

   def center(self):
      qr = self.frameGeometry()
      cp = QtWidgets.QDesktopWidget().availableGeometry().center()
      qr.moveCenter(cp)
      self.move(qr.topLeft())

   exportSounding = pyqtSignal(str)

   def export(self):
     self.exportSounding.emit(self.lineEdit.text())

   def update_plot(self):
     self.drawButton.setEnabled(False)
     z,th,p,qv = self.SoundingTabs.currentWidget().plot()
     z2,u,v = self.WindTabs.currentWidget().plot()
     self.Sounding.plot_sounding(z,th,p,qv,u,v)
     self.Hodograph.plot_hodograph(z2,u,v)

   @pyqtSlot()
   def enableButton(self):
     self.drawButton.setEnabled(True)



"""Class to read GRaDS style CM1 model output data

"""
#TODO exception throwing?  Error handling, support the u,v,w,stats files

import numpy as np
import mmap

class CM1(object):
   """Class that implements reading CM1 model data

   :param path: path to CM1 data files
   :param datasetname: the CM1 data files basename
   
   """
   nx   = 0
   """gridpoints in the x direction"""
   ny   = 0
   """gridpoints in the y direction"""
   nz   = 0
   """gridpoints in the z direction"""
   nv   = 0
   """number of variables in the dataset"""
   nt   = 0
   """number of time levels in the dataset"""
   dt   = 0
   """timestep between time levels"""
   dimX = 0
   """X grid dimension (1D)"""
   dimY = 0
   """Y grid dimension (1D)"""
   dimZ = 0
   """Z grid dimension (1D)"""
   dimT = 0
   """T grid dimension (1D)"""
   vars = 0
   n2d  = 0
   recl = 0
   dsetname = ''
   path = ''
   mem_file = 0
   dat_file = 0

#-------------------------------------------------------

   def __init__(self,__path,__datasetname):
      
      self.path = __path
      self.dsetname = __datasetname

      #We need to open the control file and read the dataset metadata 

      #create filename.  This assumes cm1 created files and the scalar file at that.
      ctl_filename = self.path + '/' + self.dsetname + '_s.ctl'

      # open the control file
      with open(ctl_filename, 'r') as ctl_file:
         print('Opened '+ ctl_filename+ ' for reading')
         #read ctl file
         self.parsectl(ctl_file)
      #close ctl file
      ctl_file.closed
      print(ctl_filename+ ' closed')

#-------------------------------------------------------
#  This function currently only works for cm1 domains with stretched x/y/z grids.

   def parsectl(self,ctl_file):

      print('  Parsing control file')
      line = ctl_file.readline()
      while line:
         # This finds the line that describes the X dimension
        if line.startswith('xdef'):
          nX = [int(s) for s in line.split() if s.isdigit()].pop()
          print('    Found X with {0} values'.format(nX))
          X = []
          for i in range(nX):
            line = ctl_file.readline()
            X.append(float(line))
          if nX != len(X):
            print('Error: Expected {0} values, read {1} values, exiting!'.format(nX,len(X)))
            quit()
          self.dimX = np.array(X)
          self.nx = nX

         # This finds the line that describes the Y dimension
        if line.startswith('ydef'):
          nY = [int(s) for s in line.split() if s.isdigit()].pop()
          print('    Found Y with {0} values'.format(nY))
          Y = []
          for i in range(nY):
            line = ctl_file.readline()
            Y.append(float(line))
          if nY != len(Y):
            print('Error: Expected {0} values, read {1} values, exiting!'.format(nY,len(Y)))
            quit()
          self.dimY = np.array(Y)
          self.ny = nY

         # This finds the line that describes the Z dimension
        if line.startswith('zdef'):
          nZ = [int(s) for s in line.split() if s.isdigit()].pop()
          print('    Found Z with {0} values'.format(nZ))
          Z = []
          for i in range(nZ):
            line = ctl_file.readline()
            Z.append(float(line))
          if nZ != len(Z):
            print('Error: Expected {0} values, read {1} values, exiting!'.format(nZ,len(Z)))
            quit()
          self.dimZ = np.array(Z)
          self.nz = nZ

         # This finds the line that describes the time dimension
        if line.startswith('tdef'):
          nT = [s for s in line.split()]
          self.nt = int(nT[1])
          self.dt = float([s for s in nT[4].split('YR')][0])
          print('    Found T with {0} steps of {1} s'.format(self.nt, self.dt))
          T = []
          for i in range(self.nt):
            T.append(float(i*self.dt))
            self.dimT = np.array(T)

         # This finds the line that begins the variable declaration
        if line.startswith('vars'):
          nV = [int(s) for s in line.split() if s.isdigit()].pop()
          print('    Found {0} variables'.format(nV))
          V = []
          for i in range(nV):
            line = ctl_file.readline()
            newV = [i+1] #varid

            vdef = ([s for s in line.split(None,3)])
            newV.append(vdef[0]) # varname
            newV.append(int(vdef[1])) #levels
            if int(vdef[1]) == 0:
              self.n2d = self.n2d + 1
            newV.append(vdef[3].rstrip()) # description
            newV = tuple(newV)

            V.append(newV)
          if nV != len(V):
            print('Error: Expected {0} values, read {1} values, exiting!'.format(nV,len(V)))
            quit()
          self.vars = V
          self.nv = nV
          #eat the endvars line
          line = ctl_file.readline()

        line = ctl_file.readline()
      
      self.recl = self.nx * self.ny * 4

#-------------------------------------------------------

   def getVarByName(self,varname):
      idx = [i for i in self.vars if i[1] == varname]
      return {'id': int(idx[0][0]), 'name': idx[0][1], 'nlevs': int(idx[0][2]), 'desc': idx[0][3]}

#-------------------------------------------------------
# This function reads count number of bytes sequentially from location loc
# from the memory region mm (which may be a memory mapped file).

   def readData(self, mm, loc, count):
      
      # Reads data starting at loc, length count
      bytes = 0
      data = b''

      # seek to location
      mm.seek(loc)

      #read until we have all of the bytes
      while (bytes < count):
         newdata = mm.read(count-bytes)
         if newdata==0:
            print('Error reading memmapped data')
            #Throw
         bytes += len(newdata)
         data += newdata

      # did we get all of the bytes?
      if count != len(data):
         print('Error reading memmapped data')
	       #throw

      return data

#-------------------------------------------------------
#  This function reads one record from the datafile.  Each
#  record is nx*ny*4 bytes and is one xy-slice of the domain

   def read3dXYSlice(self, memmap, varid, level):

      # Calculate the record number and then mult by the record
      # length to calculate the offset into the file the record begins at

      idx = (self.n2d+(varid-1-self.n2d)*(self.nz))+level # record number
      loc = idx*self.recl

      # Read the data starting at location loc, length recl

      data = self.readData(memmap, loc, self.recl)

      # did we get all of the bytes?
      if self.recl != len(data):
         print('Error reading memmapped data')
	 #throw

      # convert data to numpy array of floats and shape it into i*j
      newarray = np.fromstring(data, dtype=np.float32, count=int(self.recl/4))
      newarray = newarray.reshape((self.nx, self.ny)).T
      return newarray

#-------------------------------------------------------
# Reads a single 3d variable from the datafile

   def read3d(self, time, varname):
      """Reads a 3D variable from the dataset

      :param time: the timelevel to read
      :param varname: the variable name to read
      :reaturns: 3D array containing the variable at the time
      """
      var = self.getVarByName(varname)

      # open dat file 
      dat_filename = self.path + '/' + self.dsetname + '_{0:06d}_s.dat'.format(int(time))

      print('  Opening {0} for reading'.format(dat_filename))
      print('    Reading {0}({1}) at time {2} s'.format(varname, var['id'], time))

      # open the data file
      with open(dat_filename, 'rb') as dat_file:

        # memmap the data file
        memmap = mmap.mmap(dat_file.fileno(), 0, prot=mmap.PROT_READ)

	      # init array for the variable
        data = np.empty((self.nx, self.ny, self.nz), dtype=np.float32)

        # Read the data one record at a time until we have the full domain volume
        for z in range(self.nz):
          data[:,:,z] = self.read3dXYSlice(memmap, var['id'], z)
      
      #close memory map and the data file
      memmap.close()
      dat_file.closed
      print('  {0} closed'.format(dat_filename))

      return data

#-------------------------------------------------------
# These 3 functions seperate the opening, reading and closing
#  actions so that mutliple variables can be read from the same
#  datafile with only 1 open/close action.
      
   def read3dMultStart(self,time):

      # create datafile name
      dat_filename = self.path + '/' + self.dsetname + '_{0:06d}_s.dat'.format(int(time))
      print('  Opening {0} for reading'.format(dat_filename))
      print('    Reading time {0} s'.format(time))
   
      # open and memory map the file
      self.dat_file = open(dat_filename, 'rb') 
      self.map_file = mmap.mmap(self.dat_file.fileno(), 0, prot=mmap.PROT_READ)
      
#-------------------------------------------------------

   def read3dMultStop(self):

      # close the memory map and the data file
      self.map_file.close()
      self.dat_file.closed
      print('  File closed')

#-------------------------------------------------------

   def read3dMult(self, varname):

      # get the variable information
      var = self.getVarByName(varname)
      print('    Reading var {0}({1})'.format(varname,var['id']))

      #initialize the array to hold the data
      data = np.empty((self.nx, self.ny, self.nz), dtype=np.float32)

      # read the data one slice at a time until we have the full volume
      for z in range(self.nz):
         data[:,:,z] = self.read3dXYSlice(self.map_file, var['id'], z)

      return data

#-------------------------------------------------------

# def get var by id, get varid by name
# def var exists?



#!/usr/bin/env python
"""
.. module:: pymeteo.thermo
   :platform: Unix, Windows
   :synopsis: Thermodynamic routines

.. moduleauthor:: Casey Webster <casey.webster@gmail.com>

"""
#todo
# complete constants and function library

#constants

import math
import numpy as np
from pymeteo.constants import *
import pymeteo.interp

def T(theta,p):
    """Convert Potential Temperature :math:`\\theta` to Temperature

    :parameter theta: Potential temperature (K)
    :parameter p: Pressure (Pa)
    :returns: Temperature (K)
    """
    return theta * (p00/p)**-kappa_d

def theta(T,p):
	return T * (p00/p)**kappa_d

def es(T):
	#return 611.2 * np.exp((Lv(T)/Rv)*((1./T00)-(1./T)))
        return 611.2 * np.exp(17.67*(T-T00)/(T-29.65))

def esi(T):
	#return 611.2 * np.exp((Lv(T)/Rv)*((1./T00)-(1./T)))
        return 611.2 * np.exp(21.8745584*(T-T00)/(T-7.66))

def w_vs(T,pd):
	return epsilon * (es(T)/pd)

def theta_v(th, qv):
   return th * (1. + 0.61*qv)

def Td(p, qv):
    old_settings = np.seterr(all='ignore')
    el = np.log((qv/epsilon)*p/100./(1. + (qv/epsilon)))
    Td = T00 + (243.5*el-440.8)/(19.48-el)
    np.seterr(**old_settings)
    return Td

def Lv(T):
#TODO: Temp dependance
	return L

def p_from_pressure_altitude(pz, T):
   #TODO: T00 should be mean T in layer
   p = p00*np.exp((g/(Rd*T00))*(-pz))
   return p

def th_e(p, t, td, qv):
   if (td-t) >= -0.1:
      tlcl = t
   else:
      tlcl = 56.0 + ((td-56.0)**(-1) + 0.00125*math.log(t/td) )**(-1)

   th_e = t * ((100000./p) ** (0.2854*(1.0-0.28*qv)))*math.exp(((3376./tlcl)-2.54)*qv*(1.0+0.81*qv))

   return th_e

def q_vl(p, t):
   _es = es(t)
   q_vl = epsilon*_es/(p-_es)
   return q_vl

def q_vi(p, t):
   _es = esi(t)
   q_vi = epsilon*_es/(p-_es)
   return q_vi

# 16 Nov 2011 class notes -- PSU Meteo 531 Jerry Harrington
def dTdz_moist(T,p):
  pd = p - es(T)
  num = 1. + ((Lv(T) * w_vs(T,pd))/(Rd*T))
  den = 1. + ((Lv(T)**2 * w_vs(T,pd))/(cpd * Rv * T**2))
  return (-g/cpd)*(num/den)

def dTdp_moist(T,p):
	return dTdz_moist(T,p) * -((Rd*T)/(p*g))

def Twb(z,p,th,qv,z0):
   # returns wet bulb (deg C) for parcel from height z0

   if (len(z) != len(p) != len(_th) != len(qv)):
      raise Exception('Bounds of z, T, Td do not match')

   th_0 = pymeteo.interp.linear(z,th,z0)
   qv0 = pymeteo.interp.linear(z,qv,z0)
   p0 = pymeteo.interp.interp_pressure(p,z,z0)
   t0 = T(th_0, p0)
   td0 = Td(p0, qv0)
   es0 = es(td0)

   #print(qv0, p0, t0, td0)
   if np.isnan(td0):
       return np.nan

   residual = 0.0005
   not_converged = 1
   i = 0
   Twb = td0-T00
   sign = 1
   prevsign = 1
   incr = 10
   while not_converged:
      Ewguess = es(Twb+T00)
      Eguess = Ewguess/100. - p0/100. * (t0-T00-Twb) * 0.00066 * (1. + (0.00115 * (Twb)))
      delta_e = es0/100. - Eguess

      if abs(delta_e) <= residual:
         not_converged = 0
         break

      if delta_e < 0.:
         sign = -1
      else:
         sign = 1

      if sign != prevsign:
         prevsign = sign
         incr /= 10.

#      print Twb, delta_e, incr, sign
      Twb += incr * float(sign)
      i+=1

      if i > 100:
         print('Failed to converge')
         Twb = 0.
         break

   return Twb

def CAPE(z, p, t, q, parcel):

   #TODO: TOPS, LImax|500|300

   # T in K
   # P in Pa
   # q in kg/kg

   if (len(z) != len(p) != len(t) != len(q)):
      raise Exception('Bounds of z, T, Td do not match')

   ml_depth = .500  # for option of mixed layer parcel.
   pinc = 100. # Pa

   adiabat = 1

   debuglevel = 0

   nk = len(z)

   # goal: pi, td, th, thv
   # have: z, p, t, q
   pi  = np.empty(nk, np.float32)
   td  = np.empty(nk, np.float32)
   th  = np.empty(nk, np.float32)
   thv = np.empty(nk, np.float32)

   pi = (p*rp00)**rddcp
   td = Td(p,q)
   th = t/pi
   thv = th * (1. + reps * q)/(1. + q)

   #TODO: calc z from hydrostatic instead of using provided levels?

   # source parcel

   if (parcel == 1):
      kmax = 0

   elif (parcel == 2):
      # use most unstable parcel

      if (p[0] < 50000.):
         kmax = 0
         maxthe = th_e(p[0], t[0], td[0], q[0])
      else:
         maxthe = 0.
         for k in range(nk):
            if p[k] < 50000.:
               break
            the = th_e(p[k], t[k], td[k], q[k])
            if (the > maxthe):
               maxthe = the
               kmax = k
      if (debuglevel >= 100):
         print('  kmax,maxthe = {0}, {1}'.format(kmax,maxthe))

   elif (parcel == 3):
      # mixed layer

      if ((z[1]-z[0]) > ml_depth):
         # second level is above mixed layer
         avgth = th[0]
         avgqv = q[0]
         kmax = 0

      elif (z[nk-1] < ml_depth):
         # all levels are in the mixed layer
         #print nk-1, z[nk-1], ml_depth
         avgth = th[nk-1]
         avgqv = q[nk-1]
         kmax = nk-1

      else:
         # calculate mixed layer properties

         avgth = 0.
         avgqv = 0.
         k = 1
         if (debuglevel >= 100):
            print('  ml_depth = {0}'.format(ml_depth))
            print('  k,z,th,q = {0}, {1}, {2}, {3}'.format(0, z[0], th[0], q[0]))

         while ((z[k] <= ml_depth) and (k < nk)):
            if (debuglevel >= 100):
               print('  k,z,th,q = {0}, {1}, {2}, {3}'.format(k, z[k], th[k], q[k]))

            avgth += 0.5 * (z[k] - z[k-1])*(th[k] + th[k-1])
            avgqv += 0.5 * (z[k] - z[k-1])*(q[k] + q[k-1])

            k += 1

         th2 = th[k-1]+(th[k]-th[k-1])*(ml_depth-z[k-1])/(z[k]-z[k-1])
         qv2 = q[k-1]+(q[k]-q[k-1])*(ml_depth-z[k-1])/(z[k]-z[k-1])

         if (debuglevel >= 100):
               print('  k,z,th,q = {0}, {1}, {2}, {3}'.format(999, ml_depth, th2, qv2))

         avgth += 0.5*(ml_depth-z[k-1])*(th2+th[k-1])
         avgqv += 0.5*(ml_depth-z[k-1])*(qv2+q[k-1])


         if (debuglevel >= 100):
               print('  k,z,th,q = {0}, {1}, {2}, {3}'.format(k, z[k], th[k], q[k]))

         avgth /= ml_depth
         avgqv /= ml_depth

         kmax = 0

      if (debuglevel >= 100):
         print('  avgth, avgqv = {0}, {1}'.format(avgth, avgqv))

   # define parcel properties
   narea = 0.

   if (parcel == 1) or (parcel ==2) :
      k = kmax
      th2 = th[kmax]
      pi2 = pi[kmax]
      p2  = p[kmax]
      t2  = t[kmax]
      thv2= thv[kmax]
      qv2 = q[kmax]
      b2 = 0.
   elif (parcel == 3):
      k = kmax
      th2 = avgth
      qv2 = avgqv
      thv2= th2*(1.+reps*qv2)/(1.+qv2)
      pi2 = pi[kmax]
      p2  = p[kmax]
      t2  = th2*pi2
      b2  = g*(thv2-thv[kmax])/thv[kmax]

   p0 = p2
   t0 = t2
   qv0 = qv2

   ql2 = 0.
   qi2 = 0.
   qt = qv2

   cape = 0.
   cin = 0.
   lfc = 0.

   doit = True
   cloud = False
   if (adiabat == 1) or (adiabat == 2):
      ice = False
   else:
      ice = True

   the = th_e(p2, t2, t2, qv2)
   th_p_e = the
   if (debuglevel >= 100):
      print('th_e = {0}'.format(the))

   pt = np.empty(nk, np.float32)
   pb = np.empty(nk, np.float32)
   pc = np.empty(nk, np.float32)
   pn = np.empty(nk, np.float32)
   ptv = np.empty(nk, np.float32)
   ptd = np.empty(nk, np.float32)
   pqv = np.empty(nk, np.float32)
   pql = np.empty(nk, np.float32)

   pt[k] = t2
   if cloud:
      ptd[k] = t2
   else:
      ptd[k] = Td(p2, qv2)
   ptv[k] = t2*(1.+reps*qv2)/(1.+qv2)
   pb[k] = 0.
   pqv[k] = qv2
   pql[k] = 0.

   zlcl = -1.
   zlfc = -1.
   zel = -1.
   ztops = -1.
   ptops = 0.
   plcl = 0.
   plfc = 0.
   pel = 0.
   max_li = 40.

   # Parcel Ascent starts here!

   if (debuglevel >= 100):
      print('  Start loop:')
      print('  p2,th2,qv2 = {0}, {1}, {2}'.format(p2, th2, qv2))

   while ((doit) and (k < (nk-1))):
      k = k+1
      b1 = b2

      dp = p[k-1]-p[k]
      nloop = 1
      if (dp >= pinc):
         nloop = 1 + int(dp/pinc)
         dp = dp/float(nloop)

      for n in range(nloop):

         p1 = p2
         t1 = t2
         pi1 = pi2
         th1 = th2
         qv1 = qv2
         ql1 = ql2
         qi1 = qi2
         thv1 = thv2

         p2 = p2 - dp
         pi2 = (p2*rp00)**rddcp

         thlast = th1
         i = 0
         not_converged = True

         while (not_converged):

            i += 1
            t2 = thlast*pi2
            if ice:
               fliq = max(min((t2-233.15)/(T00-233.15),1.),0.)
               fice = 1.-fliq
            else:
               fliq = 1.
               fice = 0.

            qv2 = min( qt, fliq*q_vl(p2,t2) + fice*q_vi(p2,t2))
            qi2 = max( fice*(qt-qv2), 0.)
            ql2 = max( qt-qv2-qi2, 0.)

            tbar = 0.5*(t1+t2)
            qvbar = 0.5*(qv1+qv2)
            qlbar = 0.5*(ql1+ql2)
            qibar = 0.5*(qi1+qi2)

            lhv = lv1-lv2*tbar
            lhs = ls1-ls2*tbar
            lhf = lhs - lhv

            rm = Rd + Rv*qvbar
            cpm = cp + cpv*qvbar + cpl*qlbar + cpi*qibar
            th2 = th1*math.exp( lhv*(ql2-ql1)/(cpm*tbar) + lhs*(qi2-qi1)/(cpm*tbar) + (rm/cpm - Rd/cp)*math.log(p2/p1))

            if i > 90:
               print(i,th2,thlast,th2-thlast)

            if i > 100:
               raise ValueError('Error: lack of convergence, stopping iteration')

            if abs(th2-thlast) > converge:
               thlast += 0.3*(th2-thlast)
            else:
               not_converged = False

         # end not_converged

         # pressure increment complete
         if (ql2 >= 1.0e-10):
            cloud = True

         if cloud and (zlcl < 0.):
            zlcl = z[k-1] + (z[k]-z[k-1])*float(n)/float(nloop)
            plcl = p[k-1] + (p[k]-p[k-1])*float(n)/float(nloop)

         if (adiabat == 1) or (adiabat == 3):
            #pseudoadiabat
            qt = qv2
            ql2 = 0.
            qi2 = 0.
         elif (adiabat <= 0) or (adiabat >= 5):
            raise ValueError('Adiabat our of range')

      # end nloop

      thv2 = th2*(1.+reps*qv2)/(1.+qv2+ql2+qi2)
      b2 = g * (thv2 - thv[k])/thv[k]
      dz = -cpdg*0.5*(thv[k]+thv[k-1])*(pi[k]-pi[k-1])

      if (thv[k] - thv2) < max_li:
         max_li = thv[k] - thv2

      if ( zlcl > 0.) and (zlfc < 0.) and (b2 > 0.):
         if (b1 > 0.):
           zlfc = zlcl
           plfc = plcl
         else:
           zlfc = z[k-1]+(z[k]-z[k-1])*(0.-b1)/(b2-b1)
           plfc = p[k-1]+(p[k]-p[k-1])*(0.-b1)/(b2-b1)

      if ( zlfc > 0.) and (zel < 0.) and (b2 < 0.):
        zel = z[k-1]+(z[k]-z[k-1])*(0.-b1)/(b2-b1)
        pel = p[k-1]+(p[k]-p[k-1])*(0.-b1)/(b2-b1)


      the = th_e(p2, t2, t2, qv2)

      pt[k] = t2
      if (cloud):
         ptd[k] = t2
      else:
         ptd[k] = Td(p2,qv2)
      ptv[k] = t2*(1.+reps*qv2)/(1.+qv2)
      pb[k] = b2
      pqv[k] = qv2
      pql[k] = ql2

      # Cape and cin contributions
      if (b2 >= 0.) and (b1 < 0.):
         # First positive area
         ps = p[k-1]+(p[k]-p[k-1])*(0.-b1)/(b2-b1)
         frac = b2/(b2-b1)
         parea = 0.5*b2*dz*frac
         narea -= 0.5*b1*dz*(1.-frac)
         if (debuglevel >= 200):
            print('        b1,b2 = {0},{1}'.format(b1,b2))
            print('        p1,ps,p2 = {0},{1},{2}'.format(p[k-1],ps,p[k]))
            print('        frac = {0}'.format(frac))
            print('        parea = {0}'.format(parea))
            print('        narea = {0}'.format(narea))

         cin += narea
         narea = 0.0
      elif (b2 < 0.) and (b1 > 0.):
         # first negative area
         ps = p[k-1]+(p[k]-p[k-1])*(0.-b1)/(b2-b1)
         frac = b1/(b1-b2)
         parea = 0.5*b1*dz*frac
         narea = -0.5*b2*dz*(1.-frac)
         if (debuglevel >= 200):
            print('        b1,b2 = {0},{1}'.format(b1,b2))
            print('        p1,ps,p2 = {0},{1},{2}'.format(p[k-1],ps,p[k]))
            print('        frac = {0}'.format(frac))
            print('        parea = {0}'.format(parea))
            print('        narea = {0}'.format(narea))

      elif (b2 < 0.):
         #still neg
         parea = 0.
         narea -= 0.5*dz*(b1+b2)
      else:
         parea = 0.5*dz*(b1+b2)
         narea = 0.

      cape += max(0., parea)
      pc[k] = cape
      #print parea,narea,cape,cin,zel,ztops,z[k],(b1-b2),(narea-cape)
      if (debuglevel >= 200):
         print(p2, b1, b2, cape, cin, cloud)

      if (narea >= cape) and (zel > 0.) and (ztops < 0.):
         #print('Tops')
         ztops = z[k-1]+(z[k]-z[k-1])*(b1-b2)/(narea-cape)
         ptops = p[k-1]+(p[k]-p[k-1])*(b1-b2)/(narea-cape)

      if (p[k] <= 10000.) and (b2 < 0.):
         doit = False

   # end parcel ascent loop

   li500 = 0
   li300 = 0

   lev = 50000.
   k = pymeteo.interp.interp_height(z, p, lev)
   t_p_lev = pymeteo.interp.linear(z, ptv, k)
   t_e_lev = T(pymeteo.interp.linear(z, thv, k),lev)
   li500 = t_e_lev - t_p_lev

   lev = 30000
   k = pymeteo.interp.interp_height(z, p, lev)
   t_p_lev = pymeteo.interp.linear(z, ptv, k)
   t_e_lev = T(pymeteo.interp.linear(z, thv, k),lev)
   li300 = t_e_lev - t_p_lev

   #print('CAPE = {0}'.format(cape))
   dict = { 'lfc' : zlfc,
            'lcl' : zlcl,
            'el'  : zel,
            'lfcprs' : plfc,
            'lclprs' : plcl,
            'elprs'  : pel,
            'ztops'  : ztops,
            'ptops'  : ptops,
            'lfcth' : 0,
            'lclth' : 0,
            'elth'  : 0,
            'zlevs' : z,
            't_p'  : pt,
            'tv_p'  : ptv,
            'thv_env'  : thv,
            'pp'    : p,
            'theta_e': th_p_e,
            'cape'  : cape,
            'cin'   : cin,
            'max_li': max_li,
            'li500' : li500,
            'li300' : li300,
            'prs'   : p[kmax]
          }

   return dict




from __future__ import print_function
from PyQt5 import QtWidgets, QtGui, QtCore
from PyQt5.QtCore import pyqtSlot, pyqtSignal
import numpy as np

from pymeteo import skewt
from pymeteo import dynamics

# widgets -> skewt wind sounding etc

class StatWidget(QtWidgets.QFrame):

  statsdone = pyqtSignal()

  def __init__(self, parent=None):
    super(StatWidget, self).__init__(parent)
    self.initUI()
    self.z = 0
    self.u = 0
    self.v = 0
    self.th = 0
    self.p = 0
    self.qv = 0
    self.SStats = 0
    self.HStats = 0

  def initUI(self):
    self.statsBox = QtWidgets.QTextBrowser(self)
    #self.updateBtn = QtGui.QPushButton("Update", self)
    #self.updateBtn.clicked.connect(self.calcStats)

    layout = QtWidgets.QVBoxLayout()
    layout.addWidget(self.statsBox)
    #layout.addWidget(self.updateBtn)
    self.setLayout(layout)

  def calcStats(self):
    self.statsBox.clear()
    if (self.SStats == 0):
      self.statsBox.append("No Sounding Plotted")  
    else:
      pcl, mupcl, mlpcl = skewt.calc_sounding_stats(self.z, self.th, self.p, self.qv)
      self.statsBox.append("Var\t\tSFC\tML\tMU")
      self.statsBox.append("CAPE\tJ/kg\t{0:.0f}\t{1:.0f}\t{2:.0f}".format(pcl['cape'],mlpcl['cape'],mupcl['cape']))
      self.statsBox.append("CIN\tJ/kg\t{0:.1f}\t{1:.1f}\t{2:.1f}".format(pcl['cin'],mlpcl['cin'],mupcl['cin']))
      self.statsBox.append("LI_max\t\t{0:.1f}\t{1:.1f}\t{2:.1f}".format(pcl['max_li'],mlpcl['max_li'],mupcl['max_li']))
      self.statsBox.append("LI_500\t\t{0:.1f}\t{1:.1f}\t{2:.1f}".format(pcl['li500'],mlpcl['li500'],mupcl['li500']))
      self.statsBox.append("LI_300\t\t{0:.1f}\t{1:.1f}\t{2:.1f}".format(pcl['li300'],mlpcl['li300'],mupcl['li300']))
      self.statsBox.append("TOPS\tm\t{0:.0f}\t{1:.0f}\t{2:.0f}".format(pcl['ztops'],mlpcl['ztops'],mupcl['ztops']))

      self.statsBox.append("LCL\tmb (m)\t{0:.0f} ({1:.0f})\t{2:.0f} ({3:.0f})\t{4:.0f} ({5:.0f})".format(pcl['lclprs']/100.,pcl['lcl'],mlpcl['lclprs']/100.,mlpcl['lcl'],mupcl['lclprs']/100.,mupcl['lcl']))
      self.statsBox.append("LCL\tmb (m)\t{0:.0f} ({1:.0f})\t{2:.0f} ({3:.0f})\t{4:.0f} ({5:.0f})".format(pcl['lfcprs']/100.,pcl['lfc'],mlpcl['lfcprs']/100.,mlpcl['lfc'],mupcl['lfcprs']/100.,mupcl['lfc']))

      self.statsBox.append("PRS\tmb\t{0:.0f}\t{1:.0f}\t{2:.0f}".format(pcl['prs']/100.,mlpcl['prs']/100.,mupcl['prs']/100.))
      self.statsBox.append("")
      shear = skewt.calc_hodograph_stats(self.z, self.u, self.v)
      cth,cr = dynamics.uv_to_deg(shear['bunkers'][0],shear['bunkers'][1])
      self.statsBox.append("Storm motion (left mover): {0:.0f} deg {1:5.2f} m/s".format(cth,cr))
      self.statsBox.append("Storm motion (left mover): u={0:5.2f} v={1:5.2f} m/s".format(shear['bunkers'][0],shear['bunkers'][1]))
      self.statsBox.append("")
      self.statsBox.append("Layer\tSRH\tERH")
      self.statsBox.append("0-1 km\t{0:.0f}\t{1:.0f}".format(shear['srh01'],shear['erh01']))
      self.statsBox.append("0-3 km\t{0:.0f}\t{1:.0f}".format(shear['srh03'],shear['erh03']))
      self.statsBox.append("")
      self.statsBox.append("Layer\tShear Vector")
      self.statsBox.append("0-1 km\t{0:.0f} deg, {1:5.2f} m/s".format(*shear['s01']))
      self.statsBox.append("0-3 km\t{0:.0f} deg, {1:5.2f} m/s".format(*shear['s03']))
      self.statsBox.append("0-6 km\t{0:.0f} deg, {1:5.2f} m/s".format(*shear['s06']))

      self.statsdone.emit()

  @pyqtSlot(np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray)
  def SStat_values(self, z,th,p,qv,u,v):
    self.z = z
    self.th = th
    self.p = p
    self.qv = qv
    self.u = u
    self.v = v
    self.SStats = 1
    self.calcStats()

  @pyqtSlot(str)
  def output_sounding(self, filename):
    print("would output to file {0}".format(filename))
    f = open(filename, 'w')
    for k in range(len(self.z)):
      if (k==0):
        print("{0:8.2f}\t{1:8.6f}\t{2:8.6f}".format(self.p[0]/100., self.th[0], self.qv[0]*1000.), file=f)
      else:
        print("{0:8.2f}\t{1:8.6f}\t{2:8.6f}\t{3:8.6f}\t{4:8.6f}".format(self.z[k], self.th[k], self.qv[k]*1000., self.u[k], self.v[k]), file=f)
    f.close() 



from PyQt5 import QtWidgets, QtGui, QtCore
from PyQt5.QtCore import pyqtSignal
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as \
    FigureCanvas
from matplotlib.backends.backend_qt5agg import NavigationToolbar2QT as \
    NavigationToolbar
import matplotlib.pyplot as plt
import numpy as np

from pymeteo.cm1.soundings import WK82
from pymeteo import skewt

# widgets -> skewt wind sounding etc


class PlotWidget(QtWidgets.QFrame):

    SStats = pyqtSignal(np.ndarray, np.ndarray, np.ndarray, np.ndarray,
                        np.ndarray, np.ndarray)

    def __init__(self, parent=None):
        super(PlotWidget, self).__init__(parent)
        self.initUI()

    def initUI(self):
        self.figure = plt.figure()
        self.canvas = FigureCanvas(self.figure)
        self.toolbar = NavigationToolbar(self.canvas, self)

        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.toolbar)
        layout.addWidget(self.canvas)
        self.setLayout(layout)

    def plot_sounding_axes(self):
        self.figure.clear()
        ax = self.figure.add_axes([0.05, 0.05, 0.90, 0.945])
        # ax = self.figure.add_axes([0.005,0.05,0.985,0.945])
        skewt.set_fontscalefactor(4)
        skewt.plot_sounding_axes(ax)
        self.canvas.draw()

    def plot_hodograph_axes(self):
        ax = self.figure.add_axes([0.005, 0.005, 0.985, 0.985])
        skewt.plot_hodo_axes(ax)
        self.canvas.draw()

    def plot_sounding(self, z, th, p, qv, u, v):
        self.figure.clear()
        # ax = self.figure.add_axes([0.005,0.05,0.985,0.945])
        ax = self.figure.add_axes([0.05, 0.05, 0.90, 0.945])
        skewt.plot_sounding_axes(ax)
        skewt.plot_sounding(ax, z, th, p, qv, u, v)
        self.canvas.draw()
        # Send data to stats widget
        self.SStats.emit(z, th, p, qv, u, v)

    def plot_hodograph(self, z, u, v):
        self.figure.clear()
        ax = self.figure.add_axes([0.005, 0.05, 0.985, 0.945])
        skewt.plot_hodo_axes(ax)
        skewt.plot_hodograph(ax, z, u, v)
        self.canvas.draw()



"""This module provides constants used in the rest of the package

"""
# some constants

missingval = -99999999.
m2km = 0.001
km2m = 1000.
gravity = 9.81
maxparcels = 99999
L = 2.501e6    # latent heat of vaporization
Rd = 287.04         # gas constant dry air
Rv = 461.5              # gas constant water vapor
epsilon = Rd/Rv
cp = 1005.7              # what about cpd vs cpv
cpd = 1005.7             # what about cpd vs cpv
cpv = 1875.0
cpl = 4190.0
cpi = 2118.636
cv = 718.
g = 9.81
p00 = 100000.   # reference pressure
T00 = 273.15
xlv = L
xls = 2836017.0

# Derivced values

lv1 = xlv+(cpl-cpv)*T00
lv2 = cpl - cpv
ls1 = xls+(cpi-cpv)*T00
ls2 = cpi - cpv

kappa = (cp-cv)/cp
kappa_d = Rd/cp
rp00 = 1./p00
reps = Rv/Rd
eps = epsilon
rddcp = kappa_d
cpdrd = cp/Rd
cpdg = cp/g

converge = 0.0002




from PyQt5 import QtWidgets, QtGui, QtCore
from PyQt5.QtCore import pyqtSignal, pyqtSlot
from pymeteo.data import acars
import pymeteo.cm1.PlotWidget as PW
import pymeteo.cm1.StatWidget as SW

class MainWindow(QtWidgets.QWidget):

    def __init__(self):
        super(MainWindow, self).__init__()
        self.initUI()

    def initUI(self):
        self.splitter_2 = QtWidgets.QSplitter(QtCore.Qt.Horizontal)

        self.LVert_widget = QtWidgets.QWidget(self.splitter_2)
        self.verticalLayout = QtWidgets.QVBoxLayout(self.LVert_widget)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.label = QtWidgets.QLabel('Output file:',self.LVert_widget)
        self.horizontalLayout.addWidget(self.label)
        self.lineEdit = QtWidgets.QLineEdit(self.LVert_widget)
        self.horizontalLayout.addWidget(self.lineEdit)
        self.pushButton = QtWidgets.QPushButton("Export", self.LVert_widget)
        self.horizontalLayout.addWidget(self.pushButton)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.line = QtWidgets.QFrame(self.LVert_widget)
        self.line.setFrameShape(QtWidgets.QFrame.HLine)
        self.line.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.verticalLayout.addWidget(self.line)

        self.label2 = QtWidgets.QLabel('Select ACARS data')
        self.verticalLayout.addWidget(self.label2)

        self.acarsFiles = QtWidgets.QComboBox(self)
        self.verticalLayout.addWidget(self.acarsFiles)

        self.line_2 = QtWidgets.QFrame(self.LVert_widget)
        self.line_2.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_2.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.verticalLayout.addWidget(self.line_2)

        self.label3 = QtWidgets.QLabel('Select Profile')
        self.verticalLayout.addWidget(self.label3)
        self.profiles = QtWidgets.QComboBox(self)
        self.verticalLayout.addWidget(self.profiles)

        self.verticalLayout.addStretch(1)

        self.drawButton = QtWidgets.QPushButton("Update Plot", self.LVert_widget)
        self.quitButton = QtWidgets.QPushButton("Quit", self.LVert_widget)
        self.verticalLayout.addWidget(self.drawButton)
        self.verticalLayout.addWidget(self.quitButton)
        self.splitter = QtWidgets.QSplitter(self.splitter_2)
        self.splitter.setOrientation(QtCore.Qt.Vertical)
        self.Hodograph = PW.PlotWidget(self.splitter)
        self.SoundingInfo = SW.StatWidget(self.splitter)
        self.Sounding = PW.PlotWidget(self.splitter_2)

        datasets = acars.getAvailableDatasets()
        self.acarsFiles.addItems(datasets)
        self.acarsFiles.activated[str].connect(self.processDataset)

        self.profiles.activated.connect(self.displayProfile)

        self.Sounding.SStats.connect(self.SoundingInfo.SStat_values)

        self.pushButton.clicked.connect(self.export)
        self.exportSounding.connect(self.SoundingInfo.output_sounding)

        self.drawButton.clicked.connect(self.update_plot)
        self.quitButton.clicked.connect(self.close)

        self.hbox = QtWidgets.QHBoxLayout(self)
        self.hbox.addWidget(self.splitter_2)
        self.setLayout(self.hbox)

        self.resize(1200,768)

        #TODO: size widgets initially. This doesnt work
        self.LVert_widget.resize(250,500)
        self.Hodograph.resize(250,250)
        self.SoundingInfo.resize(250,250)
        self.Sounding.resize(250,500)


        self.center()
        self.setWindowTitle('ACARS Profile Plotting Tool')

        self.show()
        self.Sounding.plot_sounding_axes()
        self.Hodograph.plot_hodograph_axes()
        self.SoundingInfo.calcStats()

        self.SoundingInfo.statsdone.connect(self.enableButton)

    def center(self):
        qr = self.frameGeometry()
        cp = QtWidgets.QDesktopWidget().availableGeometry().center()
        qr.moveCenter(cp)
        self.move(qr.topLeft())

    def profileDesc(self,profile):
        profileDir = "UP" if profile["flag"] == 1 else "DOWN"
        return "{0}: {1}({2}) {3}".format(profile["i"], profile["airport"], profileDir, profile["time"])

    def processDataset(self, acarsFile):
        print("[+] Getting Data")
        self.profiles.clear()
        self.data = acars.processDataSet(acars.getDataSet(acarsFile)) 
        for profile in self.data:
            desc = self.profileDesc(profile)
            print("[-] Adding profile {0}".format(desc))
            self.profiles.addItem(desc)

    def displayProfile(self, index):
        print("[+] Displaying profile {0}".format(index))
        profile = self.data[index]
        desc = self.profileDesc(profile)
        print("[-] {0}".format(desc))
        self.Sounding.plot_sounding(
            profile["z"],
            profile["th"],
            profile["p"],
            profile["qv"],
            profile["u"],
            profile["v"]
        )
        self.Hodograph.plot_hodograph(
            profile["z"],
            profile["u"],
            profile["v"]
        )


    exportSounding = pyqtSignal(str)

    def export(self):
       self.exportSounding.emit(self.lineEdit.text())

    def update_plot(self):
       self.drawButton.setEnabled(False)


    @pyqtSlot()
    def enableButton(self):
       self.drawButton.setEnabled(True)    



"""Skew-T/Log-P plotting, meteorology and interfacing with CM1

This package provides routines for plotting Skew-T/Log-P diagrams
and working with CM1 output.  For contributions, please submit
pull requests to https://github.com/cwebster2/pymeteo. Please report
bugs using the github issue tracker at https://github.com/cwebster2/pymeteo/issues.

"""
__version__="1.0.1"



# wrf.py
import numpy as np

######
# The following code is adapted from NCL which is copyright:
#
# Copyright © 2007-2015 University Corporation for Atmospheric Research (UCAR)
#
# and subject to the following license
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Neither the names of NCAR's Computational and Information Systems Laboratory,
# the University Corporation for Atmospheric Research, nor the names of its
# contributors may be used to endorse or promote products derived from this
# Software without specific prior written permission.  Redistributions of source
# code must retain the above copyright notices, this list of conditions, and the
# disclaimer below.  Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions, and the disclaimer below in the
# documentation and/or other materials provided with the distribution.  THIS
#
# SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
# INCLUDING, BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
# PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE CONTRIBUTORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
# CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.
#
# https://www.ncl.ucar.edu/Download/NCL_source_license.shtml

def ll_to_ij(map_proj, truelat1, truelat2, stand_lon, dx, dy, ref_lat, ref_lon, lat, lon):
    
    if (map_proj == 6):
        #TODO: implement map_proj 6
        #pole_lat = f.getncattr('POLE_LAT')
        #pole_lon = f.getncattr('POLE_LON')
        latinc = ( dy * 360. ) / 2.0 / np.pi / 6.37e6
        loninc = ( dx * 360. ) / 2.0 / np.pi / 6.37e6
    else:
        pole_lat = 90.0
        pole_lon = 0.0
        latinc = 0.0
        loninc = 0.0

    re = 6.37e6                
    rebydx = re / dx
    radperdeg = np.pi/180.0
    degperrad = 180.0/np.pi
    hemi = 1.0
    if (truelat1 < 0.0):
        hemi = -1.0

    if (map_proj == 3): # mercator
        clain = np.cos(radperdeg * truelat1)
        dlon = dx / ( re * clain)
        rsw = 0.0
        if (ref_lat != 0):
                rsw = np.log(np.tan(0.5*((ref_lat+90.0) * radperdeg))) / dlon
        deltalon = lon - ref_lon
        if (deltalon < -180.0):
                deltalon = deltalon + 360.0
        if (deltalon > 180.0):
                deltalon = deltalon - 360.0
        i = 0 + (deltalon / (dlon * degperrad))
        j = 0 + np.log(np.tan(0.5*((lat+90.0)*radperdeg)))/dlon - rsw

    elif (map_proj == 2): # polar-stereo
        reflon = stand_lon + 90.0
        scale_top = 1.0 + hemi*np.sin(truelat1*radperdeg)
        ala1 = ref_lat*radperdeg
        rsw = rebydx*np.cos(ala1)*scale_top/(1.0+hemi*np.sin(ala1))
        alo1 = (ref_lon - reflon)*radperdeg
        polei = 1.0 - rsw * np.cos(alo1)
        polej = 1.0 - hemi*rsw*np.sin(alo1)
        ala = lat*radperdeg
        rm = rebydx*np.cos(ala)*scale_top / (1.0+hemi*np.sin(ala))
        alo = (lon-reflon)*radperdeg
        i = polei + rm*np.cos(alo)
        j = polej + hemi*rm*np.sin(alo)

    elif (map_proj == 1): # lambert
        if (np.abs(truelat2) > 90.0):
                truelat2 = truelat1
        if (np.abs(truelat1-truelat2) > 0.1):
                cone = (np.log(np.cos(truelat1*radperdeg))-
                        np.log(np.cos(truelat2*radperdeg))) /\
                        (np.log(np.tan((90.0-np.abs(truelat1))*radperdeg*0.50))-
                         np.log(np.tan((90.0-np.abs(truelat2))*radperdeg*0.50)))
        else:
                cone = np.sin(np.abs(truelat1)*radperdeg)

        deltalon1 = ref_lon - stand_lon
        if (deltalon1 > 180.0):
                deltalon = deltalon1 - 360.0
        if (deltalon1 < -180.0):
                deltalon = deltalon1 + 360.0
        tl1r = truelat1*radperdeg
        ctl1r = np.cos(tl1r)

        rsw = rebydx*ctl1r/cone* (np.tan((90.0*hemi-ref_lat)*radperdeg/2.0) /
              np.tan((90.0*hemi-truelat1)*radperdeg/2.0))**cone

        arg = cone * (deltalon1*radperdeg)
        polei = hemi*1.0 - hemi*rsw*np.sin(arg)
        polej = hemi*1.0 + rsw*np.cos(arg)

        deltalon = lon - stand_lon
        if (deltalon > 180.0):
                deltalon = deltalon - 360.0
        if (deltalon < -180.0):
                deltalon = deltalon + 360.0

        rm = rebydx*ctl1r/cone* (np.tan((90.0*hemi-lat)*radperdeg/2.0)/
                                 np.tan((90.0*hemi-truelat1)*radperdeg/2.0))**cone
        arg = cone * (deltalon*radperdeg)
        i = polei + hemi*rm*np.sin(arg)
        j = polej - rm*np.cos(arg)

        i = hemi*i
        j = hemi*j

    #TODO: implement map projection 6
    #elif (map_proj == 6):
    else:
        print("Unsupported map projection")
        return

    i = int(i+0.5)
    j = int(j+0.5)

    return (i,j)

########################################################



from PyQt5 import QtWidgets, QtGui, QtCore

class OptionsWidget(QtWidgets.QFrame):

  def __init__(self):
    super(OptionsWidget,self).__init__()
 
  def initUI(self, name, v):
    self.qlbl = QtWidgets.QLabel(name)
    self.grid = QtWidgets.QGridLayout()
    self.grid.addWidget(self.qlbl,1,0,1,3)

    nv = len(v)
    self.variables = {}

    for var in range(nv):
      self.variables[v[var][0]] = self.addOption(var, *v[var])

    self.grid.setSpacing(5)
    self.setLayout(self.grid)

  def addOption(self, pos, name, defval, units):
    label = QtWidgets.QLabel(name)
    var   = QtWidgets.QLineEdit(defval, self)
    label2= QtWidgets.QLabel(units)

    self.grid.addWidget(label,pos+2,0)
    self.grid.addWidget(var,pos+2,1)
    self.grid.addWidget(label2,pos+2,2)

    return var
    
  def getOption(self, name):
    return float(self.variables[name].text())




"""
.. module: pymeteo.uwyo
   :platform: Unix, Windows
   :synopsis: University of Wyoming sounding data helper functions

.. moduleauthor:: Casey Webster <casey.webster@gmail.com>

This module provides functions for working with University of Wyoming sounding
data.  This includes functions for getting data from files and from the University
of Wyoming website at http://weather.uwyo.edu/ and a function for preparing that data
for SkewT plotting.

Getting Data
++++++++++++

* :py:func:`fetch_from_file` -- loads sounding data from a file
* :py:func:`fetch_from_web` -- loads sounding data from website

Preparing Data
++++++++++++++

* :py:func:`transform_and_check_data` -- quality checks and converts units of data

Module Reference
++++++++++++++++
"""

import numpy as np
import io
import datetime
import pymeteo.dynamics as dynamics
try:
    # For Python 3.0 and later
    from urllib import request
except ImportError:
    # Fall back to Python 2's urllib2
    import urllib2 as request

def fetch_from_file(filename):
    """Load Uinversity of Wyoming sound data from a file

    :param filename: The filename containing sounding data
    :type filename: str
    :returns: pressure, heights, mixing ration, wind direction, wind speed and potential temperature

    This loads sounding data from a file and returns the data columns needed
    for SkewT plotting.  The data format is the same as copy/pasting data from
    the uwyo website.
    """
    with open(filename, 'r') as f:
        title = f.readline()
        skiprows = 7
        if "Obs" not in title:
            title = filename
            skiprows = 5

    p, z, qv, wind_dir, wind_speed, th = np.genfromtxt(filename, unpack=True, skip_header=skiprows,
                                                           delimiter=(7,7,7,7,7,7,7,7,7,7,7,7),
                                                           usecols=(0,1,5,6,7,8))
    return (title, p, z, qv, wind_dir, wind_speed, th)

def fetch_from_web(date, station):
    """Load Uinversity of Wyoming sound data from the uwyo website

    :param date: Time and date of requested sounding data
    :type date: datetime
    :param station: The station ID for sounding data
    :type station: int
    :returns: pressure, heights, mixing ration, wind direction, wind speed and potential temperature

    This loads sounding data from the web and returns the data columns needed
    for SkewT plotting.  The date is rounded to the preceding 12-hourly observation at 12Z or 00Z.
    """
    # http://weather.uwyo.edu/cgi-bin/sounding
    # region=naconf
    # TYPE=TEXT-LIST
    # YEAR=2016
    # MONTH=01
    # FROM=0212
    # TO=0212
    # STNM=72251

    year = date.year
    month = date.month
    day = date.day
    hour = date.hour
    if hour < 12:
        hour = 00
    else:
        hour = 12
    print(year, month, day, hour, station)
    base_url = "http://weather.uwyo.edu/cgi-bin/sounding"
    url = "{0}?TYPE=TEXT%3ALIST&YEAR={1}&MONTH={2:02d}&FROM={3:02d}{4:02d}&TO={3:02d}{4:02d}&STNM={5}".format(
           base_url, year, month, day, hour, station)
    #print(url)
    data=[]
    bulkdata=[]
    urlreq = request.Request(url) #, method='GET')

    try:
        with request.urlopen(urlreq) as f:
            bulkdata = str(f.read()).split(r'\n')[0:-1]
            for i in range(len(bulkdata)):
                bulkdata[i] = bulkdata[i][0:]

        if f.status != 200:
            print("Error fetching data");
            return

    except AttributeError:
        print("python2")
        f = request.urlopen(urlreq)
        bulkdata = f.read().split('\n')
        for i in range(len(bulkdata)):
            bulkdata[i] = bulkdata[i][0:]

    parse_state = 'start'
    for i in range(len(bulkdata)):
        if parse_state == 'start':
            if '<H2>' not in bulkdata[i]:
                continue
            title = (bulkdata[i][4:-5])
            parse_state = 'data'
        elif parse_state == 'data':
            if '</PRE>' in bulkdata[i]:
                parse_state = 'finished'
                continue
            if '<PRE>' in bulkdata[i]:
                continue
            data.append(bulkdata[i])
        elif parse_state == 'finished':
            pass
    data = "\n".join(data)

    if (data == ""):
        raise ValueError("ERROR: No sounding data found in data returned from the server");

    p, z, qv, wind_dir, wind_speed, th = np.genfromtxt(io.BytesIO(data.encode()), unpack=True,
                                                       skip_header=5,
                                                       delimiter=7,
                                                       usecols=(0,1,5,6,7,8))

    return (title,p,z,qv,wind_dir, wind_speed, th)


def transform_and_check_data(p, z, qv, wind_dir, wind_speed, th):
    """Quality check and convert units of Uinversity of Wyoming sound data

    :param p: pressure
    :param z: heights
    :param qv: mixing ration
    :param wind_dir: wind direction
    :param wind_speed: wind speed
    :param th: potential temperature
    :returns: pressure, heights, mixing ration, u wind, v wind and potential temperature

    This ingests data loaded from file owr web and culls invalid rows (NaN in certain colums)
    and quality checks the data.  Units are converted from the native uwyo data to the
    module standard SI units.
    """
    # clean up NaNs and mark invalid rows for deletion
    nk = len(z)
    for k in np.arange(nk):
        delete_rows = []
        if np.isnan(p[k]):
            delete_rows.append(k)
        if np.isnan(qv[k]):
            qv[k] = 0
        if np.isnan(wind_speed[k]):
            wind_speed[k] = wind_speed[k-1]
        if np.isnan(wind_dir[k]):
            wind_dir[k] = wind_dir[k-1]

    # delete invalid rows
    p = np.delete(p,delete_rows)
    z = np.delete(z,delete_rows)
    qv = np.delete(qv,delete_rows)
    wind_dir = np.delete(wind_dir, delete_rows)
    wind_speed = np.delete(wind_speed, delete_rows)
    th = np.delete(th, delete_rows)

    # convert ingested units to our package standard units
    nk = len(z)
    p = p * 100. # hPa to Pa
    qv = qv / 1000. # g/kg to kg/kg
    wind_speed = wind_speed * 0.51444  # kts to m/s

    # convert wind direction,speed to u,v components
    u = np.empty(nk, np.float32)
    v = np.empty(nk, np.float32)
    for k in np.arange(nk):
        u[k], v[k] = dynamics.wind_deg_to_uv(wind_dir[k], wind_speed[k])

    p[np.isnan(p)] = 0
    qv[np.isnan(qv)] = 0
    u[np.isnan(u)] = 0
    v[np.isnan(v)] = 0
    #reutrn data
    return (p, z, qv, u, v, th)



#!/usr/bin/env python
"""This module provides routine thermodynamic functions

"""
#todo
# complete constants and function library

#constants

import math
import numpy as np
from pymeteo.constants import *
import pymeteo.interp

# Mean wind in the layer [zmin,zmax]
def avg_wind(_u, _v, _z, zmin, zmax):
#TODO: np.average !
#REPLACED BY mean_wind below
        u = 0
        v = 0
        n = 0
        for i in np.arange(0,len(_z),1):
                if (_z[i] > zmin and _z[i] < zmax):
                        u += _u[i]
                        v += _v[i]
                        n += 1
        if (n == 0):
            return 0
        return (u/n, v/n)

def storm_motion_rasmussen(_u,_v,_z):
        #rasmussen (1984)
        # calculate storm motion
        u_0_500 = avg_wind(_u,_v,_z, 0., 500.)
        u_4km = avg_wind(_u,_v,_z, 3500.,4500.)

        dist_60pct = math.sqrt((u_4km[0]-u_0_500[0])**2 + (u_4km[1]-u_0_500[1])**2) * 0.6
        du = u_4km[0]-u_0_500[0]
        dv = u_4km[1]-u_0_500[1]
        theta = math.atan(dv/du)
        u60 = dist_60pct * math.cos(theta) + u_0_500[0]
        v60 = dist_60pct * math.sin(theta) + u_0_500[1]
        theta -= math.pi/2.
        dist = 8.7
        u_cr = dist * math.cos(theta) + u60
        v_cr = dist * math.sin(theta) + v60
        theta += math.pi
        u_cl = dist * math.cos(theta) + u60
        v_cl = dist * math.sin(theta) + v60

        return u_cr,v_cr,u_cl,v_cl

def storm_motion_bunkers(_u,_v,_z):
        #bunkers (2000)
        # calculate storm motion
        u_0_500 = avg_wind(_u,_v,_z, 0., 500.)
        u_0_6km = avg_wind(_u,_v,_z, 0., 6000.)
        du = u_0_6km[0] - u_0_500[0]
        dv = u_0_6km[1] - u_0_500[1]
        theta = math.atan(dv/du) - math.pi/2.
        dist = 7.5
        u_cr = dist * math.cos(theta) + u_0_6km[0]
        v_cr = dist * math.sin(theta) + u_0_6km[1]
        theta += math.pi
        u_cl = dist * math.cos(theta) + u_0_6km[0]
        v_cl = dist * math.sin(theta) + u_0_6km[1]

        return u_cr,v_cr,u_cl,v_cl

def circulation(u, v, w, x, y, z):

  n = len(x)

	# local variables
  uavg = np.zeros((n), np.float32)
  vavg = np.zeros((n), np.float32)
  wavg = np.zeros((n), np.float32)

  # initialize returned variables
  vdotdl1 = np.empty((n), np.float32)
  vdotdl2 = np.empty((n), np.float32)
  C = 0.0

  if (u == missingval).any() or \
     (v == missingval).any() or \
     (w == missingval).any() or \
     (x == missingval).any() or \
     (y == missingval).any() or \
     (z == missingval).any():

     del uavg, vavg, wavg, dx, dy, dz, vdotdl1, vdotdl2
     return (missingval, 0, 0)

        # calculate wind and dl around the circuit
  uavg[0:n-1] = 0.5 * (u[0:n-1] + u[1:n])
  vavg[0:n-1] = 0.5 * (v[0:n-1] + v[1:n])
  wavg[0:n-1] = 0.5 * (w[0:n-1] + w[1:n])

  uavg[n-1] = 0.5*(u[0] + u[n-1])
  vavg[n-1] = 0.5*(v[0] + v[n-1])
  wavg[n-1] = 0.5*(w[0] + w[n-1])

        # ediff1d returns an array of the difference between
	# each element of the passed array, which is to say,
	# it returns the delta of each array element.  perfect.
  dx = np.ediff1d(x, to_end=x[0]-x[n-1])
  dy = np.ediff1d(y, to_end=y[0]-y[n-1])
  dz = np.ediff1d(z, to_end=z[0]-z[n-1])

        # assumes clockwise parcels
  C = np.sum(-uavg*dx - vavg*dy - wavg*dz) 
  vdotdl1 = - uavg*dx - vavg*dy - wavg*dz
  vdotdl2 = vdotdl1 / (dx**2 + dy**2 + dz**2)**0.5
  vdotdl1 = vdotdl1 * km2m

        
  C = C * km2m  # C now in units m2/s
        
  del uavg, vavg, wavg, dx, dy, dz
  return (C, vdotdl1, vdotdl2)


def integral_Bdz(th, thp, z):

  n = len(z)

  th_avg = np.empty((n), np.float32)
  thp_avg = np.empty((n), np.float32)
  dz = np.empty((n), np.float32)

  intBdz = 0.0

  if (th == missingval).any() or \
     (thp == missingval).any() or \
     (z == missingval).any():

     intBdz = missingval
     del th_avg, thp_avg, dz
     return intBdz

  th_avg[0:n-1] = 0.5 * (th[0:n-1] + th[1:n])
  thp_avg[0:n-1] = 0.5 * (thp[0:n-1] + thp[1:n])

  th_avg[n-1] = 0.5*(th[0] + th[n-1])
  thp_avg[n-1] = 0.5*(thp[0] + thp[n-1])

  dz = np.ediff1d(z, to_end=z[0]-z[n-1])

  intBdz = np.sum(-dz * gravity * thp_avg/(th_avg - thp_avg))

  del th_avg, thp_avg, dz
  intBdz = intBdz * km2m  # units now m2/s2

  return intBdz

def integral_dt(i, t):
   n = len(t)

   iavg = np.empty((n-1), np.float32)
   dt = np.empty((n-1), np.float32)

   iavg[0:n-1] = 0.5 * (i[0:n-1] + i[1:n])

   dt = np.ediff1d(t)

   integral = np.sum(iavg * dt)

   return integral


# helper functions

def uv_to_deg(u,v):
   """transforms u, v, to direction, maginutide

   :param u: u wind component
   :param v: v wind component
   :returns: wind direction and magnitude
   """
   direction = np.arctan2(u,v)*(180./np.pi)
   speed = (u**2 + v**2)**(0.5)

   return direction,speed

def wind_deg_to_uv(direction, speed):
   """Converts direction and speed into u,v wind

   :param direction: wind direction (mathmatical angle)
   :param speed: wind magnitude
   :returns: u and v wind components
   """
   u = speed * np.sin(np.pi * (direction+180.) / 180.)
   v = speed * np.cos(np.pi * (direction+180.) / 180.)

   return u,v

def shear(_u, _v, _z, zbot, ztop):
    """Calculates the shear in the layer between zbot and ztop

    :param _u: U winds (1D vector in z)
    :param _u: V winds (1D vector in z)
    :param _z: z heights (1D vector in z)
    :param zbot: Bottom of the layer
    :param ztop: Top of the layer

    """

    if zbot < _z[0]:
        zbot = _z[0]

    ubot = pymeteo.interp.linear(_z,_u, zbot)
    vbot = pymeteo.interp.linear(_z,_v, zbot)
    utop = pymeteo.interp.linear(_z,_u, ztop)
    vtop = pymeteo.interp.linear(_z,_v, ztop)

    u = utop-ubot
    v = vtop-vbot
    return u,v

def srh(_u,_v,_z, zbot, ztop, cx, cy):
    """Calculates the storm relative helicity in the layer between zbot and ztop

    :param _u: U winds (1D vector in z)
    :param _u: V winds (1D vector in z)
    :param _z: z heights (1D vector in z)
    :param zbot: Bottom of the layer
    :param ztop: Top of the layer
    :param cx: u component of storm motion
    :param cy: v component of storm motion

    """

    if zbot < _z[0]:
        zbot = _z[0]

    dz = 10.
    z = np.arange(zbot, ztop+dz, dz)
    nk = len(z)
    u = np.empty(nk, np.float32)
    v = np.empty(nk, np.float32)
    du = np.empty(nk-1, np.float32)
    dv = np.empty(nk-1, np.float32)
    uavg = np.empty(nk-1, np.float32)
    vavg = np.empty(nk-1, np.float32)

    for k in range(nk):
        u[k] = pymeteo.interp.linear(_z,_u,z[k]) 
        v[k] = pymeteo.interp.linear(_z,_v,z[k]) 

    du = np.ediff1d(u)
    dv = np.ediff1d(v)

    uavg[0:nk-1] = 0.5 * (u[0:nk-1] + u[1:nk])
    vavg[0:nk-1] = 0.5 * (v[0:nk-1] + v[1:nk])

    srh = np.sum(-(uavg-cx)*dv + (vavg-cy)*du)
    return srh

def mean_wind(_u,_v,_z, zbot, ztop):
    """Calculates the mean wind in the layer between zbot and ztop

    :param _u: U winds (1D vector in z)
    :param _u: V winds (1D vector in z)
    :param _z: z heights (1D vector in z)
    :param zbot: Bottom of the layer
    :param ztop: Top of the layer

    """

    if zbot < _z[0]:
        zbot = _z[0]

    dz = 10.
    z = np.arange(zbot, ztop+dz, dz)
    nk = len(z)
    u = np.empty(nk, np.float32)
    v = np.empty(nk, np.float32)

    for k in range(nk):
        u[k] = pymeteo.interp.linear(_z,_u,z[k])
        v[k] = pymeteo.interp.linear(_z,_v,z[k])

    uavg = np.mean(u, dtype=np.float64)
    vavg = np.mean(v, dtype=np.float64)
    return uavg, vavg

def brn(_u,_v,_z,cape):

   u06avg = mean_wind(_u,_v,_z,0.,6000.)
   u0500avg = mean_wind(_u,_v,_z,0.,500.)
   u = u06avg[0] - u0500avg[0]
   v = u06avg[1] - u0500avg[1]

   brn = cape / (0.5 * (u**2 + v**2))
   return brn



# Acars data access
# Thanks to Rich Mamrosh @ NOAA for pointing the availability of this data out to me

import re
import netCDF4
import gzip
import io
import tempfile
import os
import numpy as np
from PyQt5 import QtWidgets, QtGui, QtCore
from PyQt5.QtCore import pyqtSignal, pyqtSlot
from datetime import datetime
from pymeteo import dynamics, thermo
try:
    # For Python 3.0 and later
    from urllib import request
except ImportError:
    # Fall back to Python 2's urllib2
    import urllib2 as request

data_url = "https://madis-data.ncep.noaa.gov/madisPublic1/data/point/acars/netcdf/"
airport_ids = {}

def getAvailableDatasets():
    # crawl links at https://madis-data.ncep.noaa.gov/madisPublic1/data/point/acars/netcdf/
    req = request.Request(data_url)
    linkMatcher = re.compile(r"\"([0-9_]+\.gz)\"")
    try: 
        print("[+] Fetching list of resources available")
        with request.urlopen(req) as f:
            print("[-] Parsing list")
            data = str(f.read())
            sets = linkMatcher.findall(data)
            return sets
    except:
        print("error")

def getDataSet(set):
    req = request.Request(data_url + set)
    try:
        print("[+] Fetching dataset {0}".format(set))
        with request.urlopen(req) as f:
            compressedData = io.BytesIO(f.read()) # gzipped data
            print("[-] Decompressing response data")
            data = gzip.GzipFile(fileobj=compressedData)
            
            return data
    except:
        print("error")

def getAirportByCode(airport_id):
    print("[+] Looking up airport id '{0}'".format(airport_id))
    datfile = os.path.join(os.path.dirname(__file__), "airport_info.dat")
    if not bool(airport_ids):
        with open(datfile, "r") as f:
            for _, line in enumerate(f):
                fields = line.strip().split()
                airport_ids[int(fields[0])] = fields[1]
    return airport_ids[airport_id]

def processDataSet(data):
    print("[+] Writing data into temporary file")
    tdata = tempfile.NamedTemporaryFile()
    tdata.write(data.read())
    print("[-] Data written to {0}".format(tdata.name))
    print("[+] Opening data as NetCDF")
    d = data.read()
    with netCDF4.Dataset(tdata.name, mode='r') as nc:
        print("[-] Dataset open with")

        _z = nc["altitude"][:]
        _T = nc["temperature"][:]
        _qv = nc["waterVaporMR"][:]
        windSpeed = nc["windSpeed"][:]
        windDir = nc["windDir"][:]
        _lon = nc["longitude"][:]
        _lat = nc["latitude"][:]
        flag = nc["sounding_flag"][:]
        _airport = nc["sounding_airport_id"][:]
        time = nc["soundingSecs"][:]

        print ("[-] {0} Records".format(len(_z)))
        #conversions
        _p = thermo.p_from_pressure_altitude(_z, _T)
        _u, _v = dynamics.wind_deg_to_uv(windDir, windSpeed)
        _th = thermo.theta(_T, _p)

        # split the arrays when the flag changes sign
        splits = np.where(np.diff(time))[0]+1

        _z = np.split(_z, splits)
        _p = np.split(_p, splits)
        _th = np.split(_th, splits)
        _qv = np.split(_qv, splits)
        _u = np.split(_u, splits)
        _v = np.split(_v, splits)
        _lat = np.split(_lat, splits)
        _lon = np.split(_lon, splits)
        _airport = np.split(_airport, splits)
        time = np.split(time, splits)
        flag = np.split(flag, splits)

        print("[-] Found {0} profiles".format(len(_z)))

        #re-shape data
        outputData = []
        for i in range(len(_z)):
            ts = time[i].compressed()
            if len(ts) == 0:
                # profiles without timestamps invalid?
                continue 

            profileDir = flag[i][0]
            if (profileDir == 0):
                continue

            z = _z[i].filled()
            p = _p[i].filled()
            th = _th[i].filled()
            qv = _qv[i].filled()
            u = _u[i].filled()
            v = _v[i].filled()
            lat = _lat[i].filled()
            lon = _lon[i].filled()
            try:
               airport = getAirportByCode(_airport[i][0])
            except KeyError:
               print('Airport not found ', _airport[i][0])
               airport = "ERROR"

            profileData = {
                "i": i,
                "n": len(z),
                "z": z if profileDir > 0 else z[::-1],
                "p": p if profileDir > 0 else p[::-1],
                "th": th if profileDir > 0 else th[::-1],
                "qv": qv if profileDir > 0 else qv[::-1],
                "u": u if profileDir > 0 else u[::-1],
                "v": v if profileDir > 0 else v[::-1],
                "lat": lat if profileDir > 0 else lat[::-1],
                "lon": lon if profileDir > 0 else lon[::-1],
                "airport": airport,
                "time": datetime.utcfromtimestamp(ts.mean()).strftime("%H%MZ"),
                "flag": profileDir
            }
            outputData.append(profileData)

        return outputData



import numpy as np
from pymeteo import constants
from .. import OptionsWidget

class straight(OptionsWidget.OptionsWidget):

  def __init__(self):
    super(straight,self).__init__()
    name = 'straight (linear increase)'
    variables = [ ('z_constabv',    '6000', 'm'),
                  ('z_constblo',       '0', 'm'),
                  ('u_max',         '30.0', 'm/s'),
                  ('u_scaling',      '1.0', ''),
                  ('v_max',          '7.0', 'm/s'),
                  ('u_adjust'     ,   '0.0', 'm/s'),
                  ('v_adjust'      ,  '0.0', 'm/s') ]
    self.initUI(name, variables)
 
  def plot(self):
    # nned z, t, th, p and qv
    z = np.arange(0., 22000., 50.)
    u = np.zeros(len(z))
    v = np.zeros(len(z))

    # parameters
    zdep1 = self.getOption('z_constabv')
    zdep0 = self.getOption('z_constblo')
    umax = self.getOption('u_max')
    sf   = self.getOption('u_scaling')
    vmax = self.getOption('v_max')
    cx = self.getOption('u_adjust')
    cy = self.getOption('v_adjust')

    for k in range(len(z)):
      if (z[k] < zdep0): # constant below this height
        u[k] = 0
      elif (z[k] < zdep1): # shear
        u[k] = ((z[k]-zdep0)/(zdep1-zdep0))*umax
        u[k] = u[k]* (1 + (sf-1)*((z[k]-zdep0)/(zdep1-zdep0)))
      else: # constant section
        u[k] = umax*sf

    v[:] = vmax

    u[:] = u[:] - cx
    v[:] = v[:] - cy

    #emit
    return (z,u,v)
    




# interpolation function

import numpy as np
import pymeteo.constants
import bisect


def linear(dim, var, dval):
   # interpolates var to dval

   if len(dim) != len(var):
      raise Exception('Dimensions of dim and var do not match!')
   dim = np.ma.masked_invalid(dim)
   var = np.ma.masked_invalid(var)
   array_mask = dim.mask | var.mask
   array_x = dim.data[~array_mask]
   array_y = var.data[~array_mask]
   index_sort = np.argsort(array_x)

   return np.interp(dval, array_x[index_sort], array_y[index_sort])


def interp_height(z, p, plvl):
   #interpolates height to a pressure level

   nlevs = len(p)

   # check bounds
   if plvl > p[0]:
      return 0
   if plvl < p[nlevs-1]:
      return -1

   z0 = 0
   while p[z0] > plvl:
      z0 = z0 + 1 

   if p[z0] == plvl:
      return z[z0]

   z1 = nlevs-1
   while p[z1] < plvl:
      z1 = z1 - 1

   #interpolate to height.  
   # Code adapted from NSHARP 95 John Hart NSSFC KCMO. 

   zdiff = z[z1] - z[z0]
   pdiff = np.log( p[z0] / p[z1])
   pdist = np.log( p[z0] / plvl)
   height = z[z0] + (( pdist / pdiff) * zdiff)

   return height

def interp_pressure(p, z, zlvl):

   nlevs = len(p)
   #print('looking for {0}'.format(zlvl))

   z0 = nlevs-1
   while z[z0] > zlvl:
     z0 = z0 - 1

   #print('lower bound found idx {0} = {1}, {2}'.format(z0, z[z0], p[z0]))

   if z[z0] == zlvl:
      return p[z0]

   z1 = 0
   while z[z1] < zlvl:
     z1 = z1 + 1
     if (z1 >= nlevs):
        return
 
   #print('upper bound found idx {0} = {1}, {2}'.format(z1, z[z1], p[z1]))

   zdiff = z[z1] - z[z0]
   zdist = zlvl - z[z0]
   pdiff = np.log( p[z1] / p[z0] )
   pres = p[z0] * np.exp((zdist/zdiff) * pdiff)
   #print(pres)
   return pres




import numpy as np
from pymeteo import constants
from .. import OptionsWidget

class curved90(OptionsWidget.OptionsWidget):

  def __init__(self):
    super(curved90,self).__init__()
    name = 'curved90 (quartercircle)'
    variables = [ ('z_curve,top',    '2000', 'm'),
                  ('z_constblo',        '0', 'm'),
                  ('z_constabv',     '6000', 'm'),
                  ('u_straight,min',  '7.0', 'm/s'),
                  ('u_straight,max', '31.0', 'm/s'),
                  ('straight,scale',  '1.0', ''),
                  ('u_adjust'     ,   '0.0', 'm/s'),
                  ('v_adjust'      ,  '0.0', 'm/s') ]
    self.initUI(name, variables)
 
  def plot(self):
    # nned z, t, th, p and qv
    z = np.arange(0., 22000., 50.)
    u = np.zeros(len(z))
    v = np.zeros(len(z))

    # parameters
    zdep0 = self.getOption('z_constblo')
    zdep1 = self.getOption('z_curve,top')
    zdep2 = self.getOption('z_constabv')
    umax1 = self.getOption('u_straight,min')
    umax2 = self.getOption('u_straight,max')
    sf    = self.getOption('straight,scale')
    cx = self.getOption('u_adjust')
    cy = self.getOption('v_adjust')

    for k in range(len(z)):
      if (z[k] < zdep0): #constant section
        u[k] = 0
        v[k] = 0
      elif (z[k] < zdep1): # curvature section
        a = ((z[k]-zdep0)/(zdep1-zdep0))*(np.pi/2.)
        u[k] = umax1-umax1*np.cos(a)
        v[k] = umax1*np.sin(a)
      elif (z[k] < zdep2): # straight section
        u[k] = (umax1+(z[k]-zdep1)*(umax2-umax1)/(zdep2-zdep1))
        v[k] = umax1
        u[k] = u[k]* (1 + (sf-1)*((z[k]-zdep1)/(zdep2-zdep1)))
      else: # constant section
        u[k] = umax2*sf
        v[k] = umax1

    u[:] = u[:] - cx
    v[:] = v[:] - cy

    #emit
    return (z,u,v)
    



"""Meteorology utilities, SkewT plotting and more

This package contains some general meteorological routines 
and some specific routines to work with CM1 and WRF.  The general
routines include generation of Skew-T/Log-P plotting.  The CM1
routines include generation of input soundings for CM1 and
classes for reading CM1 output files in grads and HDF5 format.
"""

classifiers = [
"Intended Audience :: Science/Research",
"License :: OSI Approved :: BSD License",
"Topic :: Scientific/Engineering :: Atmospheric Science",
"Development Status :: 5 - Production/Stable",
"Topic :: Software Development :: Libraries :: Python Modules",
"Programming Language :: Python :: 3",
"Programming Language :: Python :: 3.10",
]

import os
import subprocess
#from distutils.core import setup
from setuptools import setup

def readme():
  with open('README.md') as f:
    return f.read()

doclines = __doc__.split("\n")

version_py = os.path.join(os.path.dirname(__file__), 'VERSION')
with open(version_py, 'r') as fh:
    version_git = open(version_py).read().strip()

setup(name='pymeteo',
      version='{ver}'.format(ver=version_git),
      description=doclines[0],
      long_description="\n".join(doclines[2:]),
      url='https://wxster.com/projects/pymeteo',
      maintainer="Casey Webster",
      maintainer_email="casey.webster@gmail.com",
      author='Casey Webster',
      author_email='casey.webster@gmail.com',
      license='BSD 3-clause',
      platforms= ["any"],
      packages=['pymeteo','pymeteo.cm1',
                'pymeteo.cm1.hodographs',
                'pymeteo.cm1.soundings',
                'pymeteo.UI', 'pymeteo.data'],
      scripts=['bin/cm1_geninit',
               'bin/plot-acars',
               'bin/skewt',
               'bin/skewt-hdf',
               'bin/skewt-blank',
               'bin/skewt-wrf'],
      package_data={'pymeteo.data': ['airport_info.dat']},
      classifiers=classifiers
)



import numpy as np
from pymeteo import constants
from pymeteo import thermo
from .. import OptionsWidget

class WK82(OptionsWidget.OptionsWidget):

  def __init__(self):
    super(WK82,self).__init__()
    name = 'Weisman-Klemp 82 Analytic Sounding'
    variables = [ ('z_tr',   '12000', 'm'),
                  ('theta_tr', '343', 'K'),
                  ('T_tr',     '213', 'K'),
                  ('th_0',     '300', 'K'),
                  ('qv_0',   '0.014', 'kg/kg'),
                  ('p_0',   '100000', 'Pa') ]
    self.initUI(name, variables)
 
  def plot(self):
    # nned z, t, th, p and qv
    z = np.arange(0., 22000., 50.)
    th = np.zeros(len(z))
    thv = np.zeros(len(z))
    p = np.zeros(len(z))
    qv = np.zeros(len(z))
    rh = np.zeros(len(z))
    pi = np.zeros(len(z))

    # parameters
    th_sfc = self.getOption('th_0')
    qv_pbl = self.getOption('qv_0')
    z_tr = self.getOption('z_tr')
    t_tr = self.getOption('T_tr')
    th_tr = self.getOption('theta_tr')
    p_sfc = self.getOption('p_0')

    pi_sfc = (p_sfc/constants.p00)**(constants.Rd/constants.cp)
    qv_sfc = thermo.q_vl(p_sfc,th_sfc*pi_sfc)
    thv_sfc = th_sfc*(1.+qv_sfc*constants.reps)/(1.+qv_sfc)

    rh[:] = 0.
    #calculate 
    # WK82 pp 506
    for k in range(len(z)):
      if (z[k] <= z_tr):
        th[k] = th_sfc + (th_tr - th_sfc)*((z[k]/z_tr)**(1.25))
        rh[k] = 1.0-0.75*((z[k]/z_tr)**1.25)
      else:
        th[k] = th_tr * np.exp((constants.gravity/(constants.cp*t_tr))*(z[k]-z_tr))
        rh[k] = 0.25

    qv[:] = 0.

    for n in range(20):
      for k in range(len(z)):
        thv[k] = th[k]*(1.+constants.reps*qv[k])/(1.+qv[k])
      pi[0] = pi_sfc-constants.gravity*z[0]/(constants.cp*0.5*(thv_sfc+thv[0]))
      for k in range(len(z)):
        if k == 0:
          continue
        pi[k] = pi[k-1]-constants.gravity*(z[k]-z[k-1])/(constants.cp*0.5*(thv[k]+thv[k-1]))
      for k in range(len(z)):
        p[k] = constants.p00*(pi[k])**(constants.cp/constants.Rd)
      #p[:] = constants.p00*(pi[:])**(constants.cp/constants.Rd)
      for k in range(len(z)):
        qv[k] = rh[k]*thermo.q_vl(p[k],th[k]*pi[k])
        if (qv[k] > qv_pbl):
          qv[k] = qv_pbl

    #emit
    return (z,th,p,qv)
